"""Tests for the observability module."""

from __future__ import annotations

import json
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, MagicMock, PropertyMock, patch

import pytest
import selenium.common.exceptions

from graftpunk.observe import NoOpObservabilityContext, ObservabilityContext
from graftpunk.observe.capture import (
    MAX_RESPONSE_BODY_SIZE,
    NodriverCaptureBackend,
    SeleniumCaptureBackend,
    _build_har_entry,
    _is_binary_mime,
    _is_text_mime,
    _mime_to_extension,
    _should_stream_to_disk,
    _wall_time_to_iso,
    create_capture_backend,
)
from graftpunk.observe.storage import ObserveStorage

# ---------------------------------------------------------------------------
# Test helpers for NodriverCaptureBackend tests
# ---------------------------------------------------------------------------


def _make_request_entry(
    url: str = "https://example.com/api",
    method: str = "GET",
    has_post_data: bool = False,
    post_data: str | None = None,
    status: int = 200,
    mime_type: str = "application/json",
) -> dict[str, Any]:
    """Build a request-map entry for NodriverCaptureBackend tests."""
    return {
        "url": url,
        "method": method,
        "headers": {},
        "post_data": post_data,
        "has_post_data": has_post_data,
        "timestamp": None,
        "response": {
            "status": status,
            "statusText": "OK",
            "headers": {},
            "mimeType": mime_type,
        },
    }


@contextmanager
def _patch_cdp_modules(
    *,
    include_runtime: bool = False,
) -> Generator[MagicMock, None, None]:
    """Patch sys.modules to provide mock CDP network (and optionally runtime) modules.

    Yields the mock network module so tests can assert on network.enable() etc.
    """
    mock_network = MagicMock()
    mock_cdp = MagicMock()
    mock_cdp.network = mock_network
    mock_nodriver = MagicMock()
    mock_nodriver.cdp = mock_cdp

    modules: dict[str, Any] = {
        "nodriver": mock_nodriver,
        "nodriver.cdp": mock_cdp,
        "nodriver.cdp.network": mock_network,
    }
    if include_runtime:
        mock_runtime = MagicMock()
        mock_cdp.runtime = mock_runtime
        modules["nodriver.cdp.runtime"] = mock_runtime

    with patch.dict("sys.modules", modules):
        yield mock_network


# ---------------------------------------------------------------------------
# ObserveStorage tests
# ---------------------------------------------------------------------------


class TestObserveStorage:
    """Tests for ObserveStorage file operations."""

    def test_creates_run_directory(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        assert storage.run_dir.exists()
        assert (storage.run_dir / "screenshots").is_dir()

    def test_save_screenshot(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        png_data = b"\x89PNG\r\n\x1a\nfake"
        path = storage.save_screenshot(1, "login-page", png_data)
        assert path.name == "001-login-page.png"
        assert path.read_bytes() == png_data

    def test_write_and_read_events(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        storage.write_event("click", {"target": "button"})
        storage.write_event("navigate", {"url": "https://example.com"})
        events = storage.read_events()
        assert len(events) == 2
        assert events[0]["event"] == "click"
        assert events[0]["target"] == "button"
        assert "timestamp" in events[0]
        assert events[1]["event"] == "navigate"

    def test_read_events_empty(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        assert storage.read_events() == []

    def test_read_events_skips_corrupt_lines(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        # Write a valid event, then corrupt data, then another valid event
        storage.write_event("click", {"target": "button"})
        with storage._events_path.open("a") as f:
            f.write("this is not valid json\n")
        storage.write_event("navigate", {"url": "https://example.com"})

        events = storage.read_events()
        assert len(events) == 2
        assert events[0]["event"] == "click"
        assert events[1]["event"] == "navigate"

    def test_init_rejects_empty_session_name(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid session_name"):
            ObserveStorage(tmp_path, "", "run-001")

    def test_init_rejects_slash_in_session_name(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid session_name"):
            ObserveStorage(tmp_path, "foo/bar", "run-001")

    def test_init_rejects_backslash_in_session_name(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid session_name"):
            ObserveStorage(tmp_path, "foo\\bar", "run-001")

    def test_init_rejects_dotdot_in_session_name(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid session_name"):
            ObserveStorage(tmp_path, "..", "run-001")

    def test_init_rejects_empty_run_id(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid run_id"):
            ObserveStorage(tmp_path, "mysession", "")

    def test_init_rejects_slash_in_run_id(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid run_id"):
            ObserveStorage(tmp_path, "mysession", "foo/bar")

    def test_init_rejects_backslash_in_run_id(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid run_id"):
            ObserveStorage(tmp_path, "mysession", "foo\\bar")

    def test_init_rejects_dotdot_in_run_id(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid run_id"):
            ObserveStorage(tmp_path, "mysession", "..")

    def test_init_rejects_space_in_session_name(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid session_name"):
            ObserveStorage(tmp_path, "foo bar", "run-001")

    def test_init_rejects_leading_dot_in_session_name(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid session_name"):
            ObserveStorage(tmp_path, ".hidden", "run-001")

    def test_init_rejects_special_chars_in_run_id(self, tmp_path: Path) -> None:
        with pytest.raises(ValueError, match="Invalid run_id"):
            ObserveStorage(tmp_path, "mysession", "run@001")

    def test_init_accepts_valid_names_with_dots_and_dashes(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "my-session.v2", "run-001.1")
        assert storage.run_dir.exists()

    def test_write_har(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        entries = [{"request": {"url": "https://example.com"}}]
        storage.write_har(entries)
        har = json.loads((storage.run_dir / "network.har").read_text())
        assert har["log"]["version"] == "1.2"
        assert har["log"]["creator"]["name"] == "graftpunk"
        assert har["log"]["entries"] == entries

    def test_write_console_logs(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        logs = [{"level": "error", "message": "oops"}]
        storage.write_console_logs(logs)
        console_path = storage.run_dir / "console.jsonl"
        lines = console_path.read_text().strip().split("\n")
        assert len(lines) == 1
        assert json.loads(lines[0])["message"] == "oops"

    def test_write_metadata(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "mysession", "run-001")
        meta = {"plugin": "mybank", "started_at": "2025-01-01T00:00:00"}
        storage.write_metadata(meta)
        loaded = json.loads((storage.run_dir / "metadata.json").read_text())
        assert loaded == meta


# ---------------------------------------------------------------------------
# ObservabilityContext tests
# ---------------------------------------------------------------------------


class TestObservabilityContext:
    """Tests for ObservabilityContext plugin-facing API."""

    def _make_context(
        self,
        tmp_path: Path,
        mode: str = "full",
        screenshot_data: bytes | None = b"\x89PNGfake",
    ) -> tuple[ObservabilityContext, ObserveStorage, MagicMock]:
        storage = ObserveStorage(tmp_path, "test", "run-001")
        capture = MagicMock()
        capture.take_screenshot_sync.return_value = screenshot_data
        ctx = ObservabilityContext(capture=capture, storage=storage, mode=mode)
        return ctx, storage, capture

    def test_screenshot_saves_file(self, tmp_path: Path) -> None:
        ctx, storage, capture = self._make_context(tmp_path)
        path = ctx.screenshot("after-login")
        assert path is not None
        assert path.exists()
        assert path.name == "001-after-login.png"
        capture.take_screenshot_sync.assert_called_once()

    def test_screenshot_increments_counter(self, tmp_path: Path) -> None:
        ctx, storage, capture = self._make_context(tmp_path)
        ctx.screenshot("first")
        path = ctx.screenshot("second")
        assert path is not None
        assert path.name == "002-second.png"

    def test_screenshot_returns_none_when_capture_fails(self, tmp_path: Path) -> None:
        ctx, storage, capture = self._make_context(tmp_path, screenshot_data=None)
        result = ctx.screenshot("fail")
        assert result is None

    def test_screenshot_noop_when_off(self, tmp_path: Path) -> None:
        ctx, storage, capture = self._make_context(tmp_path, mode="off")
        result = ctx.screenshot("ignored")
        assert result is None
        capture.take_screenshot_sync.assert_not_called()

    def test_log_writes_event(self, tmp_path: Path) -> None:
        ctx, storage, _capture = self._make_context(tmp_path)
        ctx.log("user_action", {"action": "click"})
        events = storage.read_events()
        assert len(events) == 1
        assert events[0]["event"] == "user_action"
        assert events[0]["action"] == "click"

    def test_log_noop_when_off(self, tmp_path: Path) -> None:
        ctx, storage, _capture = self._make_context(tmp_path, mode="off")
        ctx.log("ignored")
        assert storage.read_events() == []

    def test_log_defaults_data_to_empty(self, tmp_path: Path) -> None:
        ctx, storage, _capture = self._make_context(tmp_path)
        ctx.log("simple_event")
        events = storage.read_events()
        assert len(events) == 1
        assert events[0]["event"] == "simple_event"

    def test_mark_writes_mark_event(self, tmp_path: Path) -> None:
        ctx, storage, _capture = self._make_context(tmp_path)
        ctx.mark("before-submit")
        events = storage.read_events()
        assert len(events) == 1
        assert events[0]["event"] == "mark"
        assert events[0]["label"] == "before-submit"
        assert "timestamp" in events[0]

    def test_mark_noop_when_off(self, tmp_path: Path) -> None:
        ctx, storage, _capture = self._make_context(tmp_path, mode="off")
        ctx.mark("ignored")
        assert storage.read_events() == []

    @pytest.mark.asyncio
    async def test_screenshot_async_saves_file(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "test-session", "run-001")
        capture = AsyncMock()
        capture.take_screenshot = AsyncMock(return_value=b"\x89PNGfake")
        ctx = ObservabilityContext(capture=capture, storage=storage, mode="full")
        path = await ctx.screenshot_async("test-label")
        assert path is not None
        assert path.exists()
        assert b"\x89PNGfake" in path.read_bytes()

    @pytest.mark.asyncio
    async def test_screenshot_async_noop_when_off(self) -> None:
        ctx = NoOpObservabilityContext()
        result = await ctx.screenshot_async("test")
        assert result is None

    @pytest.mark.asyncio
    async def test_screenshot_async_returns_none_on_capture_failure(self, tmp_path: Path) -> None:
        storage = ObserveStorage(tmp_path, "test-session", "run-001")
        capture = AsyncMock()
        capture.take_screenshot = AsyncMock(return_value=None)
        ctx = ObservabilityContext(capture=capture, storage=storage, mode="full")
        result = await ctx.screenshot_async("fail-label")
        assert result is None


# ---------------------------------------------------------------------------
# NoOpObservabilityContext tests
# ---------------------------------------------------------------------------


class TestNoOpObservabilityContext:
    """Tests for the no-op implementation."""

    def test_screenshot_returns_none(self) -> None:
        ctx = NoOpObservabilityContext()
        assert ctx.screenshot("anything") is None

    def test_log_does_nothing(self) -> None:
        ctx = NoOpObservabilityContext()
        ctx.log("event", {"key": "value"})  # Should not raise

    def test_mark_does_nothing(self) -> None:
        ctx = NoOpObservabilityContext()
        ctx.mark("label")  # Should not raise

    def test_mode_is_off(self) -> None:
        ctx = NoOpObservabilityContext()
        assert ctx._mode == "off"


# ---------------------------------------------------------------------------
# Capture backend tests
# ---------------------------------------------------------------------------


class TestSeleniumCaptureBackend:
    """Tests for SeleniumCaptureBackend."""

    def test_start_capture(self) -> None:
        driver = MagicMock()
        backend = SeleniumCaptureBackend(driver)
        backend.start_capture()  # Should not raise

    def test_take_screenshot_sync(self) -> None:
        driver = MagicMock()
        driver.get_screenshot_as_png.return_value = b"png-data"
        backend = SeleniumCaptureBackend(driver)
        result = backend.take_screenshot_sync()
        assert result == b"png-data"

    def test_take_screenshot_sync_handles_error(self) -> None:
        driver = MagicMock()
        driver.get_screenshot_as_png.side_effect = selenium.common.exceptions.WebDriverException(
            "no browser"
        )
        backend = SeleniumCaptureBackend(driver)
        result = backend.take_screenshot_sync()
        assert result is None

    def test_stop_capture_collects_browser_logs(self) -> None:
        driver = MagicMock()
        # Performance log returns empty, browser log returns entries
        driver.get_log.side_effect = lambda log_type: (
            [] if log_type == "performance" else [{"level": "ERROR", "message": "oops"}]
        )
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()
        logs = backend.get_console_logs()
        assert {"level": "ERROR", "message": "oops"} in logs

    def test_stop_capture_handles_browser_log_error(self) -> None:
        driver = MagicMock()

        def side_effect(log_type: str) -> list:
            if log_type == "performance":
                return []
            raise selenium.common.exceptions.WebDriverException("not supported")

        driver.get_log.side_effect = side_effect
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()  # Should not raise
        # Console logs should be empty (perf log had nothing, browser log failed)
        assert backend.get_console_logs() == []

    def test_stop_capture_warns_on_perf_log_failure(self) -> None:
        """Performance log failure triggers gp_console.warn to notify the user."""
        driver = MagicMock()

        def get_log_side_effect(log_type: str) -> list:
            if log_type == "performance":
                raise selenium.common.exceptions.WebDriverException("unavailable")
            return []

        driver.get_log.side_effect = get_log_side_effect
        backend = SeleniumCaptureBackend(driver)

        with patch("graftpunk.console.warn") as mock_warn:
            backend.stop_capture()

        mock_warn.assert_called_once()
        msg = mock_warn.call_args[0][0]
        assert "Performance log collection failed" in msg

    def test_get_har_entries_empty_initially(self) -> None:
        driver = MagicMock()
        backend = SeleniumCaptureBackend(driver)
        assert backend.get_har_entries() == []

    def test_get_har_entries_returns_har_format(self) -> None:
        driver = MagicMock()
        backend = SeleniumCaptureBackend(driver)
        # Seed a request directly
        backend._request_map["r1"] = {
            "url": "https://example.com",
            "method": "GET",
            "headers": {},
            "timestamp": 1700000000.0,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "text/html",
            },
        }
        entries = backend.get_har_entries()
        assert len(entries) == 1
        assert entries[0]["request"]["method"] == "GET"
        assert entries[0]["response"]["status"] == 200

    @pytest.mark.asyncio
    async def test_take_screenshot_async_wraps_sync(self) -> None:
        driver = MagicMock()
        driver.get_screenshot_as_png.return_value = b"png-data"
        backend = SeleniumCaptureBackend(driver)
        result = await backend.take_screenshot()
        assert result == b"png-data"

    @pytest.mark.asyncio
    async def test_get_page_source_async(self) -> None:
        driver = MagicMock()
        driver.page_source = "<html>hello</html>"
        backend = SeleniumCaptureBackend(driver)
        result = await backend.get_page_source()
        assert result == "<html>hello</html>"

    @pytest.mark.asyncio
    async def test_get_page_source_async_handles_error(self) -> None:
        driver = MagicMock()
        type(driver).page_source = PropertyMock(
            side_effect=selenium.common.exceptions.WebDriverException("no browser")
        )
        backend = SeleniumCaptureBackend(driver)
        result = await backend.get_page_source()
        assert result is None


class TestSeleniumHARCapture:
    """Tests for full network capture in SeleniumCaptureBackend."""

    def _make_perf_entry(self, method: str, params: dict) -> dict:
        """Create a performance log entry in CDP format."""
        return {
            "message": json.dumps({"message": {"method": method, "params": params}}),
            "level": "INFO",
        }

    def _make_driver(
        self,
        perf_entries: list[dict] | None = None,
        browser_logs: list[dict] | None = None,
    ) -> MagicMock:
        """Create a mock driver with configurable log responses."""
        driver = MagicMock()

        def get_log(log_type: str) -> list:
            if log_type == "performance":
                return perf_entries or []
            if log_type == "browser":
                return browser_logs or []
            return []

        driver.get_log.side_effect = get_log
        return driver

    def test_stop_capture_parses_cdp_request_events(self) -> None:
        """Performance log requestWillBeSent events are parsed into _request_map."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-1",
                    "request": {
                        "url": "https://example.com/api",
                        "method": "GET",
                        "headers": {"Accept": "text/html"},
                        "hasPostData": False,
                        "postData": None,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.side_effect = Exception("no body")
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()

        assert "req-1" in backend._request_map
        data = backend._request_map["req-1"]
        assert data["url"] == "https://example.com/api"
        assert data["method"] == "GET"
        assert data["headers"] == {"Accept": "text/html"}
        assert data["timestamp"] == 1700000000.0

    def test_stop_capture_parses_cdp_response_events(self) -> None:
        """Performance log responseReceived events correlate with requests."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-2",
                    "request": {
                        "url": "https://example.com/data",
                        "method": "GET",
                        "headers": {},
                        "hasPostData": False,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-2",
                    "response": {
                        "url": "https://example.com/data",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {"Content-Type": "application/json"},
                        "mimeType": "application/json",
                    },
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.return_value = {
            "body": '{"ok": true}',
            "base64Encoded": False,
        }
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()

        data = backend._request_map["req-2"]
        assert data["response"]["status"] == 200
        assert data["response"]["mimeType"] == "application/json"

    def test_stop_capture_fetches_post_body(self) -> None:
        """POST body fetched via execute_cdp_cmd for entries with hasPostData."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-3",
                    "request": {
                        "url": "https://example.com/login",
                        "method": "POST",
                        "headers": {"Content-Type": "application/x-www-form-urlencoded"},
                        "hasPostData": True,
                        "postData": None,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-3",
                    "response": {
                        "url": "https://example.com/login",
                        "status": 302,
                        "statusText": "Found",
                        "headers": {},
                        "mimeType": "text/html",
                    },
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)

        def cdp_cmd(cmd: str, params: dict) -> dict:
            if cmd == "Network.getRequestPostData":
                return {"postData": "user=foo&pass=bar"}
            if cmd == "Network.getResponseBody":
                return {"body": "<html></html>", "base64Encoded": False}
            return {}

        driver.execute_cdp_cmd.side_effect = cdp_cmd
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()

        assert backend._request_map["req-3"]["post_data"] == "user=foo&pass=bar"

    def test_stop_capture_fetches_response_body(self) -> None:
        """Response body fetched via execute_cdp_cmd for text MIME types."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-4",
                    "request": {
                        "url": "https://example.com/api",
                        "method": "GET",
                        "headers": {},
                        "hasPostData": False,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-4",
                    "response": {
                        "url": "https://example.com/api",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {},
                        "mimeType": "application/json",
                    },
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.return_value = {
            "body": '{"result": 42}',
            "base64Encoded": False,
        }
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()

        resp = backend._request_map["req-4"]["response"]
        assert resp["body"] == '{"result": 42}'
        assert resp["bodySize"] == len(b'{"result": 42}')

    def test_stop_capture_skips_binary_response_body_without_disk(self) -> None:
        """Binary MIME without bodies_dir: body not stored."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-5",
                    "request": {
                        "url": "https://example.com/image.png",
                        "method": "GET",
                        "headers": {},
                        "hasPostData": False,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-5",
                    "response": {
                        "url": "https://example.com/image.png",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {},
                        "mimeType": "image/png",
                    },
                },
            ),
        ]
        import base64 as b64

        png_data = b"\x89PNG\r\n\x1a\nfake"
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.return_value = {
            "body": b64.b64encode(png_data).decode(),
            "base64Encoded": True,
        }
        backend = SeleniumCaptureBackend(driver, bodies_dir=None)
        backend.stop_capture()

        resp = backend._request_map["req-5"]["response"]
        # Body was fetched (bodySize set) but not stored (no bodies_dir, binary)
        assert resp["bodySize"] == len(png_data)
        assert resp.get("body") is None
        assert resp.get("_bodyFile") is None

    def test_stop_capture_streams_binary_to_disk(self, tmp_path: Path) -> None:
        """Binary MIME types written to bodies_dir."""
        bodies_dir = tmp_path / "bodies"
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-6",
                    "request": {
                        "url": "https://example.com/photo.png",
                        "method": "GET",
                        "headers": {},
                        "hasPostData": False,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-6",
                    "response": {
                        "url": "https://example.com/photo.png",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {},
                        "mimeType": "image/png",
                    },
                },
            ),
        ]
        import base64 as b64

        png_data = b"\x89PNG\r\n\x1a\nfake_image"
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.return_value = {
            "body": b64.b64encode(png_data).decode(),
            "base64Encoded": True,
        }
        backend = SeleniumCaptureBackend(driver, bodies_dir=bodies_dir)
        backend.stop_capture()

        resp = backend._request_map["req-6"]["response"]
        assert resp["_bodyFile"] == "bodies/req-6.png"
        assert (bodies_dir / "req-6.png").exists()
        assert (bodies_dir / "req-6.png").read_bytes() == png_data

    def test_stop_capture_streams_large_text_to_disk(self, tmp_path: Path) -> None:
        """Text responses over max_body_size written to bodies_dir."""
        bodies_dir = tmp_path / "bodies"
        max_size = 100
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-7",
                    "request": {
                        "url": "https://example.com/large.html",
                        "method": "GET",
                        "headers": {},
                        "hasPostData": False,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-7",
                    "response": {
                        "url": "https://example.com/large.html",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {},
                        "mimeType": "text/html",
                    },
                },
            ),
        ]
        large_body = "x" * 200
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.return_value = {
            "body": large_body,
            "base64Encoded": False,
        }
        backend = SeleniumCaptureBackend(driver, max_body_size=max_size, bodies_dir=bodies_dir)
        backend.stop_capture()

        resp = backend._request_map["req-7"]["response"]
        assert resp["_bodyFile"] == "bodies/req-7.html"
        assert (bodies_dir / "req-7.html").exists()

    def test_get_har_entries_produces_valid_har_format(self) -> None:
        """get_har_entries produces HAR 1.2 spec entries after stop_capture."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-8",
                    "request": {
                        "url": "https://example.com/page",
                        "method": "GET",
                        "headers": {"Accept": "text/html"},
                        "hasPostData": False,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-8",
                    "response": {
                        "url": "https://example.com/page",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {"Content-Type": "text/html"},
                        "mimeType": "text/html",
                    },
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.return_value = {
            "body": "<html>hello</html>",
            "base64Encoded": False,
        }
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()

        entries = backend.get_har_entries()
        assert len(entries) == 1
        entry = entries[0]
        # HAR 1.2 required fields
        assert "startedDateTime" in entry
        assert "time" in entry
        assert entry["request"]["method"] == "GET"
        assert entry["request"]["url"] == "https://example.com/page"
        assert isinstance(entry["request"]["headers"], list)
        assert entry["response"]["status"] == 200
        assert entry["response"]["content"]["text"] == "<html>hello</html>"

    def test_stop_capture_handles_evicted_request(self) -> None:
        """CDP call failure (evicted data) is handled gracefully."""
        perf_entries = [
            self._make_perf_entry(
                "Network.requestWillBeSent",
                {
                    "requestId": "req-9",
                    "request": {
                        "url": "https://example.com/evicted",
                        "method": "POST",
                        "headers": {},
                        "hasPostData": True,
                        "postData": None,
                    },
                    "wallTime": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Network.responseReceived",
                {
                    "requestId": "req-9",
                    "response": {
                        "url": "https://example.com/evicted",
                        "status": 200,
                        "statusText": "OK",
                        "headers": {},
                        "mimeType": "text/html",
                    },
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)
        driver.execute_cdp_cmd.side_effect = Exception("No resource with given identifier")
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()  # Should not raise

        # Request is still in map, just without body data
        assert "req-9" in backend._request_map
        resp = backend._request_map["req-9"]["response"]
        assert resp.get("body") is None
        assert backend._request_map["req-9"]["post_data"] is None

    def test_stop_capture_parses_console_logs(self) -> None:
        """Runtime.consoleAPICalled events are parsed from performance log."""
        perf_entries = [
            self._make_perf_entry(
                "Runtime.consoleAPICalled",
                {
                    "type": "log",
                    "args": [
                        {"type": "string", "value": "hello world"},
                    ],
                    "timestamp": 1700000000.0,
                },
            ),
            self._make_perf_entry(
                "Runtime.consoleAPICalled",
                {
                    "type": "error",
                    "args": [
                        {"type": "string", "value": "something broke"},
                    ],
                    "timestamp": 1700000001.0,
                },
            ),
        ]
        driver = self._make_driver(perf_entries=perf_entries)
        backend = SeleniumCaptureBackend(driver)
        backend.stop_capture()

        logs = backend.get_console_logs()
        # Should have the 2 CDP console entries (plus any browser log entries)
        cdp_logs = [log for log in logs if "args" in log]
        assert len(cdp_logs) == 2
        assert cdp_logs[0]["level"] == "log"
        assert cdp_logs[0]["args"] == ["hello world"]
        assert cdp_logs[1]["level"] == "error"
        assert cdp_logs[1]["args"] == ["something broke"]


class TestNodriverCaptureBackend:
    """Tests for NodriverCaptureBackend."""

    def test_start_capture(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        backend.start_capture()  # Should not raise

    def test_stop_capture(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        backend.stop_capture()  # Should not raise

    def test_take_screenshot_sync_returns_none(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        assert backend.take_screenshot_sync() is None

    def test_get_har_entries_empty_initially(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        assert backend.get_har_entries() == []

    def test_get_console_logs_empty(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        assert backend.get_console_logs() == []

    def test_init_accepts_get_tab(self) -> None:
        browser = MagicMock()
        tab = MagicMock()
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
        assert backend._tab is tab

    def test_tab_property_returns_none_without_get_tab(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        assert backend._tab is None

    @pytest.mark.asyncio
    async def test_take_screenshot_async_returns_none_without_tab(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        result = await backend.take_screenshot()
        assert result is None

    @pytest.mark.asyncio
    async def test_take_screenshot_async_returns_bytes_with_tab(self, tmp_path: Path) -> None:
        browser = MagicMock()
        tab = MagicMock()
        png_data = b"\x89PNG\r\n\x1a\nfake_screenshot"

        async def mock_save_screenshot(path: str) -> None:
            Path(path).write_bytes(png_data)

        tab.save_screenshot = mock_save_screenshot
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
        result = await backend.take_screenshot()
        assert result == png_data

    @pytest.mark.asyncio
    async def test_take_screenshot_async_handles_exception(self) -> None:
        browser = MagicMock()
        tab = MagicMock()

        async def mock_save_screenshot(path: str) -> None:
            msg = "browser crashed"
            raise RuntimeError(msg)

        tab.save_screenshot = mock_save_screenshot
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
        result = await backend.take_screenshot()
        assert result is None

    @pytest.mark.asyncio
    async def test_get_page_source_async_returns_html(self) -> None:
        browser = MagicMock()
        tab = MagicMock()
        tab.get_content = AsyncMock(return_value="<html><body>Hello</body></html>")
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
        result = await backend.get_page_source()
        assert result == "<html><body>Hello</body></html>"

    @pytest.mark.asyncio
    async def test_get_page_source_async_returns_none_without_tab(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        result = await backend.get_page_source()
        assert result is None

    @pytest.mark.asyncio
    async def test_start_capture_async_enables_network_and_runtime(self) -> None:
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock()
        tab.add_handler = MagicMock()
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
        await backend.start_capture_async()
        # network.enable() and runtime.enable()
        assert tab.send.call_count == 2
        # RequestWillBeSent, ResponseReceived, LoadingFinished, ConsoleAPICalled
        assert tab.add_handler.call_count == 4

    @pytest.mark.asyncio
    async def test_start_capture_async_passes_buffer_params(self) -> None:
        """network.enable() is called with large buffer sizes to prevent body eviction."""
        with _patch_cdp_modules(include_runtime=True) as mock_network:
            browser = MagicMock()
            tab = MagicMock()
            tab.send = AsyncMock()
            tab.add_handler = MagicMock()
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            await backend.start_capture_async()

        mock_network.enable.assert_called_once_with(
            max_total_buffer_size=100 * 1024 * 1024,
            max_resource_buffer_size=10 * 1024 * 1024,
            enable_durable_messages=True,
        )

    def test_on_response_correlates_with_request_map(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        # Create a mock ResponseReceived-like event
        mock_response = MagicMock()
        mock_response.url = "https://example.com/api"
        mock_response.status = 200
        mock_response.status_text = "OK"
        mock_response.mime_type = "application/json"
        mock_response.headers = {"Content-Type": "application/json"}

        mock_event = MagicMock()
        mock_event.response = mock_response
        mock_event.request_id = "req-1"

        backend._on_response(mock_event)
        entries = backend.get_har_entries()
        assert len(entries) == 1
        assert entries[0]["response"]["status"] == 200
        assert entries[0]["request"]["url"] == "https://example.com/api"

    def test_on_response_handles_exception_gracefully(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        mock_event = MagicMock()
        mock_event.response = None

        # Should not raise
        backend._on_response(mock_event)
        # The entry is still added (with defaults) even if response is None
        # because MagicMock returns MagicMock for attribute access

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("method", "tab_attr", "operation"),
        [
            ("take_screenshot", "save_screenshot", "screenshot"),
            ("get_page_source", "get_content", "page_source"),
        ],
    )
    async def test_browser_disconnect_warns_not_traces(
        self, method: str, tab_attr: str, operation: str
    ) -> None:
        """ConnectionError produces warning, not exception traceback."""
        browser = MagicMock()
        tab = MagicMock()
        setattr(tab, tab_attr, AsyncMock(side_effect=ConnectionRefusedError("[Errno 61]")))
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)

        with patch("graftpunk.observe.capture.LOG") as mock_log:
            result = await getattr(backend, method)()

        assert result is None
        mock_log.warning.assert_called_once_with(
            "nodriver_browser_disconnected", operation=operation
        )
        mock_log.exception.assert_not_called()


# ---------------------------------------------------------------------------
# Factory tests
# ---------------------------------------------------------------------------


class TestCreateCaptureBackend:
    """Tests for create_capture_backend factory."""

    def test_creates_selenium_backend(self) -> None:
        driver = MagicMock()
        backend = create_capture_backend("selenium", driver)
        assert isinstance(backend, SeleniumCaptureBackend)

    def test_creates_nodriver_backend(self) -> None:
        browser = MagicMock()
        backend = create_capture_backend("nodriver", browser)
        assert isinstance(backend, NodriverCaptureBackend)

    def test_creates_nodriver_backend_with_get_tab(self) -> None:
        browser = MagicMock()
        tab = MagicMock()
        get_tab = lambda: tab  # noqa: E731
        backend = create_capture_backend("nodriver", browser, get_tab=get_tab)
        assert isinstance(backend, NodriverCaptureBackend)
        assert backend._tab is tab

    def test_get_tab_ignored_for_selenium(self) -> None:
        driver = MagicMock()
        backend = create_capture_backend("selenium", driver, get_tab=lambda: MagicMock())
        assert isinstance(backend, SeleniumCaptureBackend)

    def test_raises_for_unknown_backend(self) -> None:
        driver = MagicMock()
        with pytest.raises(ValueError, match="Unknown capture backend type"):
            create_capture_backend("unknown", driver)


# ---------------------------------------------------------------------------
# build_observe_context factory tests
# ---------------------------------------------------------------------------


class TestBuildObserveContext:
    """Tests for the build_observe_context factory function."""

    def test_off_mode_returns_noop_context(self) -> None:
        """Off mode returns a NoOpObservabilityContext regardless of other args."""
        from graftpunk.observe.context import build_observe_context

        ctx = build_observe_context(
            site_name="test",
            backend_type="selenium",
            driver=MagicMock(),
            mode="off",
        )
        assert isinstance(ctx, NoOpObservabilityContext)

    def test_full_mode_with_driver_creates_context_with_capture(self) -> None:
        """Full mode with a driver creates ObservabilityContext with capture and storage."""
        from graftpunk.observe.context import build_observe_context

        mock_driver = MagicMock()

        with (
            patch("graftpunk.observe.capture.create_capture_backend") as mock_create_capture,
            patch("graftpunk.observe.storage.ObserveStorage") as mock_storage_cls,
        ):
            mock_capture = MagicMock()
            mock_create_capture.return_value = mock_capture
            mock_storage = MagicMock()
            mock_storage_cls.return_value = mock_storage

            ctx = build_observe_context(
                site_name="mysite",
                backend_type="selenium",
                driver=mock_driver,
                mode="full",
            )

        assert isinstance(ctx, ObservabilityContext)
        assert not isinstance(ctx, NoOpObservabilityContext)
        assert ctx._capture is mock_capture
        assert ctx._storage is mock_storage
        assert ctx._mode == "full"
        mock_create_capture.assert_called_once_with(
            "selenium", mock_driver, bodies_dir=mock_storage.run_dir / "bodies"
        )

    def test_full_mode_without_driver_creates_context_without_capture(self) -> None:
        """Full mode with driver=None creates context without capture."""
        from graftpunk.observe.context import build_observe_context

        with (
            patch("graftpunk.observe.storage.ObserveStorage") as mock_storage_cls,
            patch("graftpunk.observe.context.LOG") as mock_log,
        ):
            mock_storage = MagicMock()
            mock_storage_cls.return_value = mock_storage

            ctx = build_observe_context(
                site_name="mysite",
                backend_type="selenium",
                driver=None,
                mode="full",
            )

        assert isinstance(ctx, ObservabilityContext)
        assert not isinstance(ctx, NoOpObservabilityContext)
        assert ctx._capture is None
        assert ctx._storage is mock_storage
        assert ctx._mode == "full"
        mock_log.warning.assert_called_once_with(
            "observe_capture_unavailable",
            site_name="mysite",
            reason="no browser driver",
        )

    def test_off_mode_does_not_create_bodies_dir(self) -> None:
        """Off mode returns NoOpObservabilityContext without creating bodies_dir."""
        from graftpunk.observe.context import build_observe_context

        ctx = build_observe_context(
            site_name="test",
            backend_type="selenium",
            driver=MagicMock(),
            mode="off",
        )
        assert isinstance(ctx, NoOpObservabilityContext)
        # No capture backend created, so no bodies_dir interaction
        assert ctx._capture is None

    def test_full_mode_passes_bodies_dir_to_capture_backend(self, tmp_path: Path) -> None:
        """Full mode passes bodies_dir derived from storage.run_dir to capture backend."""
        from graftpunk.observe.context import build_observe_context

        mock_driver = MagicMock()

        with (
            patch("graftpunk.observe.capture.create_capture_backend") as mock_create_capture,
            patch("graftpunk.observe.storage.ObserveStorage") as mock_storage_cls,
        ):
            mock_capture = MagicMock()
            mock_create_capture.return_value = mock_capture
            mock_storage = MagicMock()
            mock_storage.run_dir = tmp_path / "test-run"
            mock_storage_cls.return_value = mock_storage

            build_observe_context(
                site_name="mysite",
                backend_type="nodriver",
                driver=mock_driver,
                mode="full",
            )

        # Verify bodies_dir is derived from run_dir
        mock_create_capture.assert_called_once_with(
            "nodriver", mock_driver, bodies_dir=tmp_path / "test-run" / "bodies"
        )


# ---------------------------------------------------------------------------
# Screenshot label sanitization tests
# ---------------------------------------------------------------------------


class TestScreenshotLabelSanitization:
    """Tests that screenshot labels are sanitized to prevent path traversal."""

    def test_path_traversal_label_sanitized(self, tmp_path: Path) -> None:
        """Labels with ../ are sanitized so path traversal does not occur."""
        storage = ObserveStorage(tmp_path, "test", "run1")
        path = storage.save_screenshot(1, "../../etc/passwd", b"png_data")
        # File must be inside the screenshots directory (no traversal)
        assert path.parent == storage.run_dir / "screenshots"
        # Slashes are replaced with dashes; the file cannot escape
        assert "/" not in path.name
        assert path.exists()
        assert path.read_bytes() == b"png_data"

    def test_spaces_in_label_sanitized(self, tmp_path: Path) -> None:
        """Spaces in labels are replaced with dashes."""
        storage = ObserveStorage(tmp_path, "test", "run1")
        path = storage.save_screenshot(1, "my screenshot", b"png_data")
        assert " " not in path.name
        assert path.exists()
        assert path.read_bytes() == b"png_data"

    def test_slashes_in_label_sanitized(self, tmp_path: Path) -> None:
        """Slashes in labels are replaced with dashes."""
        storage = ObserveStorage(tmp_path, "test", "run1")
        path = storage.save_screenshot(1, "evil/path", b"png_data")
        assert "/" not in path.name
        assert path.parent == storage.run_dir / "screenshots"
        assert path.exists()
        assert path.read_bytes() == b"png_data"


# ---------------------------------------------------------------------------
# Shared HAR helper tests
# ---------------------------------------------------------------------------


class TestSharedHARHelpers:
    """Tests for shared MIME helper functions and HAR entry builder."""

    def test_is_binary_mime_image(self) -> None:
        assert _is_binary_mime("image/png") is True
        assert _is_binary_mime("Image/JPEG") is True

    def test_is_binary_mime_application_pdf(self) -> None:
        assert _is_binary_mime("application/pdf") is True

    def test_is_binary_mime_text_html_is_false(self) -> None:
        assert _is_binary_mime("text/html") is False

    def test_is_binary_mime_application_json_is_false(self) -> None:
        assert _is_binary_mime("application/json") is False

    def test_is_text_mime_json(self) -> None:
        assert _is_text_mime("application/json") is True

    def test_is_text_mime_html(self) -> None:
        assert _is_text_mime("text/html") is True

    def test_is_text_mime_javascript(self) -> None:
        assert _is_text_mime("application/javascript") is True

    def test_is_text_mime_image_is_false(self) -> None:
        assert _is_text_mime("image/png") is False

    def test_should_stream_to_disk_binary(self) -> None:
        assert _should_stream_to_disk("image/png", 100, MAX_RESPONSE_BODY_SIZE) is True

    def test_should_stream_to_disk_large_text(self) -> None:
        large = MAX_RESPONSE_BODY_SIZE + 1
        assert _should_stream_to_disk("text/html", large, MAX_RESPONSE_BODY_SIZE) is True

    def test_should_stream_to_disk_small_text(self) -> None:
        assert _should_stream_to_disk("text/html", 100, MAX_RESPONSE_BODY_SIZE) is False

    def test_should_stream_to_disk_unknown_mime(self) -> None:
        assert _should_stream_to_disk("application/x-unknown", 100, MAX_RESPONSE_BODY_SIZE) is False

    def test_mime_to_extension_known(self) -> None:
        assert _mime_to_extension("application/json") == ".json"
        assert _mime_to_extension("text/html") == ".html"
        assert _mime_to_extension("image/png") == ".png"

    def test_mime_to_extension_with_charset(self) -> None:
        assert _mime_to_extension("application/json; charset=utf-8") == ".json"

    def test_mime_to_extension_unknown(self) -> None:
        assert _mime_to_extension("application/x-custom") == ".bin"

    def test_wall_time_to_iso_with_value(self) -> None:
        result = _wall_time_to_iso(0.0)
        assert "1970-01-01" in result

    def test_wall_time_to_iso_none_returns_current(self) -> None:
        result = _wall_time_to_iso(None)
        # Should be a valid ISO string with current date
        assert "T" in result

    def test_build_har_entry_basic(self) -> None:
        data = {
            "url": "https://example.com/api",
            "method": "GET",
            "headers": {"Accept": "application/json"},
            "timestamp": 1700000000.0,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {"Content-Type": "application/json"},
                "mimeType": "application/json",
                "bodySize": 42,
                "body": '{"ok": true}',
            },
        }
        entry = _build_har_entry(data, "req-1")
        assert entry["request"]["method"] == "GET"
        assert entry["request"]["url"] == "https://example.com/api"
        assert entry["response"]["status"] == 200
        assert entry["response"]["content"]["text"] == '{"ok": true}'
        assert entry["response"]["content"]["size"] == 42

    def test_build_har_entry_with_post_data(self) -> None:
        data = {
            "url": "https://example.com/login",
            "method": "POST",
            "headers": {"Content-Type": "application/x-www-form-urlencoded"},
            "post_data": "user=foo&pass=bar",
            "timestamp": None,
        }
        entry = _build_har_entry(data, "req-2")
        assert entry["request"]["method"] == "POST"
        assert "postData" in entry["request"]
        assert entry["request"]["postData"]["text"] == "user=foo&pass=bar"

    def test_build_har_entry_no_response(self) -> None:
        data = {"url": "https://example.com", "method": "GET", "headers": {}}
        entry = _build_har_entry(data, "req-3")
        assert entry["response"]["status"] == 0

    def test_build_har_entry_body_file(self) -> None:
        data = {
            "url": "https://example.com/image.png",
            "method": "GET",
            "headers": {},
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "image/png",
                "bodySize": 1024,
                "_bodyFile": "bodies/req-4.png",
            },
        }
        entry = _build_har_entry(data, "req-4")
        assert entry["response"]["content"]["_bodyFile"] == "bodies/req-4.png"
        assert entry["response"]["content"]["text"] is None


# ---------------------------------------------------------------------------
# Nodriver HAR capture tests
# ---------------------------------------------------------------------------


class TestNodriverHARCapture:
    """Tests for full network capture in NodriverCaptureBackend."""

    def test_on_request_stores_data(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        event = MagicMock()
        event.request_id = "req-100"
        event.request.url = "https://example.com/api"
        event.request.method = "POST"
        event.request.headers = {"Content-Type": "application/json"}
        event.request.post_data = '{"key": "value"}'
        event.request.has_post_data = True
        event.wall_time = 1700000000.0

        backend._on_request(event)

        assert "req-100" in backend._request_map
        data = backend._request_map["req-100"]
        assert data["url"] == "https://example.com/api"
        assert data["method"] == "POST"
        assert data["post_data"] == '{"key": "value"}'

    def test_on_response_correlates_with_existing_request(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        # First add a request
        req_event = MagicMock()
        req_event.request_id = "req-200"
        req_event.request.url = "https://example.com/data"
        req_event.request.method = "GET"
        req_event.request.headers = {}
        req_event.request.post_data = None
        req_event.request.has_post_data = False
        req_event.wall_time = 1700000000.0
        backend._on_request(req_event)

        # Then add response
        resp_event = MagicMock()
        resp_event.request_id = "req-200"
        resp_event.response.url = "https://example.com/data"
        resp_event.response.status = 200
        resp_event.response.status_text = "OK"
        resp_event.response.headers = {"Content-Type": "text/html"}
        resp_event.response.mime_type = "text/html"
        backend._on_response(resp_event)

        data = backend._request_map["req-200"]
        assert "response" in data
        assert data["response"]["status"] == 200
        assert data["response"]["mimeType"] == "text/html"

    def test_on_response_creates_entry_if_no_request(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        resp_event = MagicMock()
        resp_event.request_id = "orphan-1"
        resp_event.response.url = "https://example.com/orphan"
        resp_event.response.status = 304
        resp_event.response.status_text = "Not Modified"
        resp_event.response.headers = {}
        resp_event.response.mime_type = "text/html"
        backend._on_response(resp_event)

        assert "orphan-1" in backend._request_map
        data = backend._request_map["orphan-1"]
        assert data["url"] == "https://example.com/orphan"
        assert data["response"]["status"] == 304

    def test_get_har_entries_produces_har_format(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        # Simulate request + response
        req_event = MagicMock()
        req_event.request_id = "req-300"
        req_event.request.url = "https://example.com/page"
        req_event.request.method = "GET"
        req_event.request.headers = {"Accept": "text/html"}
        req_event.request.post_data = None
        req_event.request.has_post_data = False
        req_event.wall_time = 1700000000.0
        backend._on_request(req_event)

        resp_event = MagicMock()
        resp_event.request_id = "req-300"
        resp_event.response.url = "https://example.com/page"
        resp_event.response.status = 200
        resp_event.response.status_text = "OK"
        resp_event.response.headers = {"Content-Type": "text/html"}
        resp_event.response.mime_type = "text/html"
        backend._on_response(resp_event)

        entries = backend.get_har_entries()
        assert len(entries) == 1
        entry = entries[0]
        assert entry["request"]["method"] == "GET"
        assert entry["response"]["status"] == 200
        assert "startedDateTime" in entry

    @pytest.mark.asyncio
    async def test_stop_capture_async_fetches_text_body(self) -> None:
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(return_value=('{"ok": true}', False))
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)

        # Seed a request with text response
        backend._request_map["req-500"] = {
            "url": "https://example.com/api",
            "method": "GET",
            "headers": {},
            "has_post_data": False,
            "post_data": None,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "application/json",
            },
        }
        await backend.stop_capture_async()
        resp = backend._request_map["req-500"]["response"]
        assert resp["body"] == '{"ok": true}'
        assert resp["bodySize"] == len(b'{"ok": true}')

    @pytest.mark.asyncio
    async def test_stop_capture_async_fetches_post_data(self) -> None:
        browser = MagicMock()
        tab = MagicMock()

        call_count = 0

        async def mock_send(cmd: Any) -> Any:
            nonlocal call_count
            call_count += 1
            # First call is get_request_post_data, second is get_response_body
            if call_count == 1:
                return "user=foo"
            return ("<html></html>", False)

        tab.send = mock_send
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)

        backend._request_map["req-600"] = {
            "url": "https://example.com/login",
            "method": "POST",
            "headers": {},
            "has_post_data": True,
            "post_data": None,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "text/html",
            },
        }
        await backend.stop_capture_async()
        # Post data should be fetched
        assert backend._request_map["req-600"]["post_data"] == "user=foo"

    @pytest.mark.asyncio
    async def test_stop_capture_async_streams_binary_to_disk(self, tmp_path: Path) -> None:
        browser = MagicMock()
        tab = MagicMock()
        bodies_dir = tmp_path / "bodies"

        import base64

        png_data = b"\x89PNG\r\n\x1a\nfake"
        encoded = base64.b64encode(png_data).decode()
        tab.send = AsyncMock(return_value=(encoded, True))

        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab, bodies_dir=bodies_dir)
        backend._request_map["req-700"] = {
            "url": "https://example.com/image.png",
            "method": "GET",
            "headers": {},
            "has_post_data": False,
            "post_data": None,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "image/png",
            },
        }
        await backend.stop_capture_async()

        resp = backend._request_map["req-700"]["response"]
        assert resp.get("_bodyFile") == "bodies/req-700.png"
        assert (bodies_dir / "req-700.png").exists()
        assert (bodies_dir / "req-700.png").read_bytes() == png_data

    @pytest.mark.asyncio
    async def test_stop_capture_async_streams_large_text_to_disk(self, tmp_path: Path) -> None:
        browser = MagicMock()
        tab = MagicMock()
        bodies_dir = tmp_path / "bodies"
        max_size = 100  # Small max for testing

        large_body = "x" * 200
        tab.send = AsyncMock(return_value=(large_body, False))

        backend = NodriverCaptureBackend(
            browser,
            get_tab=lambda: tab,
            max_body_size=max_size,
            bodies_dir=bodies_dir,
        )
        backend._request_map["req-800"] = {
            "url": "https://example.com/large",
            "method": "GET",
            "headers": {},
            "has_post_data": False,
            "post_data": None,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "text/html",
            },
        }
        await backend.stop_capture_async()

        resp = backend._request_map["req-800"]["response"]
        assert resp.get("_bodyFile") == "bodies/req-800.html"
        assert (bodies_dir / "req-800.html").exists()

    @pytest.mark.asyncio
    async def test_stop_capture_async_returns_early_without_tab(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)
        # Should not raise
        await backend.stop_capture_async()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "exc_cls,exc_msg",
        [
            (ConnectionRefusedError, "[Errno 61] Connection refused"),
            (ConnectionError, "browser died"),
            (ConnectionResetError, "[Errno 54] Connection reset by peer"),
            (BrokenPipeError, "[Errno 32] Broken pipe"),
        ],
    )
    async def test_stop_capture_async_bails_on_connection_death(
        self, exc_cls: type, exc_msg: str
    ) -> None:
        """stop_capture_async returns immediately on connection errors."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(side_effect=exc_cls(exc_msg))

        with _patch_cdp_modules():
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            for i in range(3):
                backend._request_map[f"req-{i}"] = _make_request_entry(
                    url=f"https://example.com/page{i}",
                )
            await backend.stop_capture_async()

        # Should bail after first connection error, not try all 3
        assert tab.send.call_count == 1

    @pytest.mark.asyncio
    async def test_stop_capture_async_bails_on_post_data_connection_death(self) -> None:
        """stop_capture_async returns immediately on ConnectionError during post data fetch."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(side_effect=ConnectionError("browser died"))

        with _patch_cdp_modules():
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            backend._request_map["req-post"] = _make_request_entry(
                url="https://example.com/login",
                method="POST",
                has_post_data=True,
            )
            backend._request_map["req-get"] = _make_request_entry()
            await backend.stop_capture_async()

        # Should bail after first post_data connection error
        assert tab.send.call_count == 1

    @pytest.mark.asyncio
    async def test_stop_capture_async_logs_phase_on_disconnect(self) -> None:
        """Disconnect warning includes phase kwarg for diagnostic context."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(side_effect=ConnectionRefusedError("[Errno 61] Connection refused"))

        with _patch_cdp_modules(), patch("graftpunk.observe.capture.LOG") as mock_log:
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            backend._request_map["req-0"] = _make_request_entry()
            await backend.stop_capture_async()

        mock_log.warning.assert_called_once()
        call_kwargs = mock_log.warning.call_args[1]
        assert call_kwargs["phase"] == "response_body"
        assert call_kwargs["request_id"] == "req-0"
        assert "Connection refused" in call_kwargs["error"]
        assert call_kwargs["exc_type"] == "ConnectionRefusedError"

    @pytest.mark.asyncio
    async def test_stop_capture_async_does_not_bail_on_non_connection_oserror(self) -> None:
        """OSError subclasses like PermissionError don't trigger early bail-out."""
        browser = MagicMock()
        tab = MagicMock()
        # First call raises PermissionError (an OSError but not ConnectionError),
        # second call succeeds with a body
        tab.send = AsyncMock(
            side_effect=[
                PermissionError("Permission denied"),
                ('{"ok": true}', False),
            ]
        )

        with _patch_cdp_modules():
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            backend._request_map["req-0"] = _make_request_entry()
            backend._request_map["req-1"] = _make_request_entry(
                url="https://example.com/other",
            )
            await backend.stop_capture_async()

        # Both requests attempted — PermissionError doesn't trigger bail-out
        assert tab.send.call_count == 2

    def test_on_request_handles_exception_gracefully(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        event = MagicMock()
        event.request_id = None  # Will cause str(None) = "None" but headers access fails
        event.request = None  # This will cause AttributeError

        # Should not raise
        backend._on_request(event)

    @pytest.mark.asyncio
    async def test_on_loading_finished_eagerly_fetches_body(self) -> None:
        """LoadingFinished handler fetches response body immediately."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(return_value=('{"result": "ok"}', False))
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)

        # Seed request + response metadata
        backend._request_map["req-eager"] = {
            "url": "https://example.com/api",
            "method": "GET",
            "headers": {},
            "post_data": None,
            "has_post_data": False,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "application/json",
            },
        }

        event = MagicMock()
        event.request_id = "req-eager"
        await backend._on_loading_finished(event)

        assert "req-eager" in backend._bodies_fetched
        assert backend._request_map["req-eager"]["response"]["body"] == '{"result": "ok"}'

    @pytest.mark.asyncio
    async def test_stop_capture_skips_eagerly_fetched_bodies(self) -> None:
        """stop_capture_async skips bodies already fetched by LoadingFinished."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(return_value=("unused", False))
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)

        # Seed a request that was already eagerly fetched
        backend._request_map["req-done"] = {
            "url": "https://example.com/api",
            "method": "GET",
            "headers": {},
            "post_data": None,
            "has_post_data": False,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "application/json",
                "content": {"text": "already fetched"},
            },
        }
        backend._bodies_fetched.add("req-done")

        await backend.stop_capture_async()

        # tab.send should NOT have been called (no post_data to fetch, body already done)
        tab.send.assert_not_called()

    @pytest.mark.asyncio
    async def test_on_loading_finished_handles_failure_gracefully(self) -> None:
        """Eager fetch failure is logged but does not raise."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(side_effect=RuntimeError("CDP -32000"))
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)

        backend._request_map["req-fail"] = {
            "url": "https://example.com/api",
            "method": "GET",
            "headers": {},
            "post_data": None,
            "has_post_data": False,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "application/json",
            },
        }

        event = MagicMock()
        event.request_id = "req-fail"
        # Should not raise
        await backend._on_loading_finished(event)
        assert "req-fail" not in backend._bodies_fetched

    @pytest.mark.asyncio
    async def test_on_loading_finished_logs_exception_details(self) -> None:
        """Eager fetch failure warning includes error message and exception type."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(side_effect=RuntimeError("CDP -32000: no body available"))

        with _patch_cdp_modules(), patch("graftpunk.observe.capture.LOG") as mock_log:
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            backend._request_map["req-err"] = _make_request_entry()

            event = MagicMock()
            event.request_id = "req-err"
            await backend._on_loading_finished(event)

        mock_log.warning.assert_called_once_with(
            "nodriver_eager_body_fetch_failed",
            request_id="req-err",
            url="https://example.com/api",
            error="CDP -32000: no body available",
            exc_type="RuntimeError",
        )

    @pytest.mark.asyncio
    async def test_on_loading_finished_passes_is_update_true(self) -> None:
        """Eager fetch uses _is_update=True to skip _register_handlers() overhead."""
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock(return_value=('{"ok": true}', False))

        with _patch_cdp_modules():
            backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
            backend._request_map["req-fast"] = _make_request_entry()

            event = MagicMock()
            event.request_id = "req-fast"
            await backend._on_loading_finished(event)

        # Verify _is_update=True was passed to tab.send()
        tab.send.assert_called_once()
        _, kwargs = tab.send.call_args
        assert kwargs.get("_is_update") is True


# ---------------------------------------------------------------------------
# Nodriver console capture tests
# ---------------------------------------------------------------------------


class TestNodriverConsoleCapture:
    """Tests for console log capture in NodriverCaptureBackend."""

    def test_on_console_stores_log_entry(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        event = MagicMock()
        event.type_ = MagicMock()
        event.type_.value = "log"
        arg1 = MagicMock()
        arg1.value = "hello world"
        event.args = [arg1]
        event.timestamp = 1700000000.0

        backend._on_console(event)

        logs = backend.get_console_logs()
        assert len(logs) == 1
        assert logs[0]["level"] == "log"
        assert logs[0]["args"] == ["hello world"]
        assert logs[0]["timestamp"] == 1700000000.0

    def test_on_console_handles_missing_args(self) -> None:
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        event = MagicMock()
        event.type_ = MagicMock()
        event.type_.value = "error"
        event.args = None
        event.timestamp = None

        backend._on_console(event)

        logs = backend.get_console_logs()
        assert len(logs) == 1
        assert logs[0]["level"] == "error"
        assert logs[0]["args"] == []

    def test_on_console_handles_type_without_value_attr(self) -> None:
        """When type_ has no .value attribute, str(type_) is used."""
        browser = MagicMock()
        backend = NodriverCaptureBackend(browser)

        event = MagicMock()
        # Use a simple string-like object that has no .value
        event.type_ = "warning"
        arg1 = MagicMock()
        arg1.value = "test"
        event.args = [arg1]
        event.timestamp = 1700000000.0

        backend._on_console(event)

        logs = backend.get_console_logs()
        assert len(logs) == 1
        # str("warning") -> "warning", hasattr(str, "value") is False
        assert logs[0]["level"] == "warning"

    @pytest.mark.asyncio
    async def test_start_capture_async_enables_runtime(self) -> None:
        browser = MagicMock()
        tab = MagicMock()
        tab.send = AsyncMock()
        tab.add_handler = MagicMock()
        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab)
        await backend.start_capture_async()
        # Verify runtime.enable() was called (second send call)
        assert tab.send.call_count == 2

    @pytest.mark.asyncio
    async def test_stop_capture_binary_without_bodies_dir(self) -> None:
        """Binary MIME with no bodies_dir: body is fetched but not stored."""
        browser = MagicMock()
        tab = MagicMock()

        import base64

        png_data = b"\x89PNG\r\n\x1a\nfake"
        encoded = base64.b64encode(png_data).decode()
        tab.send = AsyncMock(return_value=(encoded, True))

        backend = NodriverCaptureBackend(browser, get_tab=lambda: tab, bodies_dir=None)
        backend._request_map["r1"] = {
            "url": "https://example.com/image.png",
            "method": "GET",
            "headers": {},
            "has_post_data": False,
            "post_data": None,
            "timestamp": None,
            "response": {
                "status": 200,
                "statusText": "OK",
                "headers": {},
                "mimeType": "image/png",
            },
        }

        await backend.stop_capture_async()

        # Body was fetched but not stored (no bodies_dir, not text)
        response = backend._request_map["r1"]["response"]
        assert response.get("body") is None
        assert response.get("_bodyFile") is None



"""Session caching and persistence using dill (enhanced pickle).

This module provides session storage functionality with pluggable backends:
- Local filesystem (default): ~/.config/graftpunk/sessions/
- Supabase (GRAFTPUNK_STORAGE_BACKEND=supabase): Supabase Storage
- S3 (GRAFTPUNK_STORAGE_BACKEND=s3): S3-compatible storage (AWS, R2, MinIO)

Backend selection is automatic based on GRAFTPUNK_STORAGE_BACKEND environment variable.

Thread Safety:
    This module uses a global cached storage backend for performance. The cache
    is NOT thread-safe. This is acceptable for the current single-threaded CLI
    usage pattern. If using graftpunk in a multi-threaded application, external
    synchronization is required when calling cache functions.
"""

import hashlib
import re
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, runtime_checkable
from urllib.parse import urlparse

import dill as pickle
import requests

from graftpunk.config import get_settings
from graftpunk.encryption import decrypt_data, encrypt_data
from graftpunk.exceptions import EncryptionError, SessionExpiredError, SessionNotFoundError
from graftpunk.logging import get_logger
from graftpunk.storage.base import SessionMetadata

if TYPE_CHECKING:
    from graftpunk.storage.base import SessionStorageBackend


@runtime_checkable
class SessionLike(Protocol):
    """Protocol for objects that can be loaded as sessions.

    This protocol defines the minimum interface expected from session objects
    returned by load_session(). It ensures type safety while allowing both
    BrowserSession and requestium.Session objects.
    """

    cookies: Any
    headers: Any


LOG = get_logger(__name__)

T = TypeVar("T")

_SESSION_NAME_RE = re.compile(r"^[a-z0-9][a-z0-9_-]*$")


def validate_session_name(name: str) -> None:
    """Validate a session name.

    Session names must be lowercase alphanumeric with hyphens/underscores,
    starting with a letter or digit. Dots are not allowed (they indicate domains).

    Raises:
        ValueError: If name is invalid.
    """
    if not name:
        raise ValueError("Session name must be non-empty")
    if "." in name:
        raise ValueError(
            f"Session name {name!r} cannot contain dots. "
            "Dots are reserved for domain matching in 'gp session clear'."
        )
    if not _SESSION_NAME_RE.match(name):
        raise ValueError(
            f"Session name {name!r} must match pattern [a-z0-9][a-z0-9_-]* "
            "(lowercase alphanumeric, hyphens, underscores)"
        )


# Global session storage backend (lazy-loaded)
_session_storage_backend: "SessionStorageBackend | None" = None


def _get_session_storage_backend() -> "SessionStorageBackend":
    """Get or create the session storage backend.

    Returns the appropriate backend based on GRAFTPUNK_STORAGE_BACKEND env var:
    - "local" (default): Returns LocalSessionStorage
    - "supabase": Returns SupabaseSessionStorage
    - "s3": Returns S3SessionStorage
    """
    global _session_storage_backend

    if _session_storage_backend is not None:
        return _session_storage_backend

    settings = get_settings()

    if settings.storage_backend == "supabase":
        # Lazy import to avoid circular imports
        from graftpunk.storage.supabase import SupabaseSessionStorage

        config = settings.get_storage_config()
        _session_storage_backend = SupabaseSessionStorage(
            url=config["url"],
            service_key=config["service_key"],
            bucket_name=config.get("bucket_name", "sessions"),
        )
    elif settings.storage_backend == "s3":
        from graftpunk.storage.s3 import S3SessionStorage

        config = settings.get_storage_config()
        _session_storage_backend = S3SessionStorage(
            bucket=config["bucket"],
            region=config.get("region"),
            endpoint_url=config.get("endpoint_url"),
            max_retries=config.get("retry_max_attempts", 5),
            base_delay=config.get("retry_base_delay", 1.0),
        )
    else:
        # Default: local filesystem backend
        from graftpunk.storage.local import LocalSessionStorage

        _session_storage_backend = LocalSessionStorage(base_dir=settings.sessions_dir)

    return _session_storage_backend


def _reset_session_storage_backend() -> None:
    """Reset the session storage backend (for testing)."""
    global _session_storage_backend
    _session_storage_backend = None


def _extract_session_metadata(session: Any, session_name: str) -> dict[str, Any]:
    """Extract metadata from a session object.

    Args:
        session: Session object
        session_name: Session name

    Returns:
        Dictionary with extracted metadata
    """
    metadata: dict[str, Any] = {"name": session_name}

    # Extract domain from current_url
    current_url = getattr(session, "current_url", None)
    if current_url:
        metadata["current_url"] = current_url
        parsed = urlparse(current_url)
        metadata["domain"] = parsed.hostname

    # Extract cookie info
    cookies = getattr(session, "cookies", None)
    if cookies:
        cookie_list = list(cookies)
        metadata["cookie_count"] = len(cookie_list)
        # Get unique domains from cookies
        domains = set()
        for cookie in cookie_list:
            if hasattr(cookie, "domain"):
                domains.add(cookie.domain)
        metadata["cookie_domains"] = sorted(domains)
    else:
        metadata["cookie_count"] = 0
        metadata["cookie_domains"] = []

    # Fallback: infer domain from cookie domains when current_url is missing
    if "domain" not in metadata and metadata.get("cookie_domains"):
        # Use the first non-empty cookie domain, stripping leading dot
        for cookie_domain in metadata["cookie_domains"]:
            if cookie_domain:
                metadata["domain"] = cookie_domain.lstrip(".")
                break

    return metadata


def get_session_metadata(name: str) -> dict[str, Any] | None:
    """Get session metadata without loading the full session.

    This is a lightweight way to check session status, timestamps, etc.
    without deserializing the encrypted pickle file.

    Args:
        name: Session name.

    Returns:
        Metadata dict if session exists, None otherwise.
    """
    backend = _get_session_storage_backend()
    metadata = backend.get_session_metadata(name)
    if metadata is None:
        return None

    # Convert SessionMetadata to dict for API compatibility
    return {
        "name": metadata.name,
        "checksum": metadata.checksum,
        "created_at": metadata.created_at.isoformat(),
        "modified_at": metadata.modified_at.isoformat(),
        "expires_at": metadata.expires_at.isoformat() if metadata.expires_at else None,
        "domain": metadata.domain,
        "current_url": metadata.current_url,
        "cookie_count": metadata.cookie_count,
        "cookie_domains": metadata.cookie_domains,
        "status": metadata.status,
    }


def cache_session(session: T, session_name: str | None = None) -> str:
    """Cache a session with metadata.

    Storage location depends on GRAFTPUNK_STORAGE_BACKEND:
    - "local" (default): sessions/{session_name}/session.pickle + metadata.json
    - "supabase": Supabase Storage bucket (file-pair pattern)
    - "s3": S3-compatible storage bucket (file-pair pattern)

    Args:
        session: Session object to cache (must be picklable).
        session_name: Optional session name. If not provided, tries to get from session.

    Returns:
        Storage location string.
    """
    if session_name is None:
        # Try to get session_name from the session object
        session_name = getattr(session, "session_name", "default")

    validate_session_name(session_name)

    backend = _get_session_storage_backend()
    settings = get_settings()

    try:
        # Pickle and encrypt session data
        LOG.info("caching_session", name=session_name, backend=settings.storage_backend)
        pickled_data = pickle.dumps(session)
        checksum = hashlib.sha256(pickled_data).hexdigest()
        encrypted_data = encrypt_data(pickled_data)

        # Extract metadata from session
        raw_metadata = _extract_session_metadata(session, session_name)
        now = datetime.now(UTC)

        # Build SessionMetadata object with configurable TTL
        metadata = SessionMetadata(
            name=session_name,
            checksum=checksum,
            created_at=now,
            modified_at=now,
            expires_at=now + timedelta(hours=settings.session_ttl_hours),
            domain=raw_metadata.get("domain"),
            current_url=raw_metadata.get("current_url"),
            cookie_count=raw_metadata.get("cookie_count", 0),
            cookie_domains=raw_metadata.get("cookie_domains", []),
            status="active",
        )

        # Save to backend
        location = backend.save_session(session_name, encrypted_data, metadata)
        LOG.info("wrote_session_to_backend", name=session_name, location=location)
        return location

    except (pickle.PickleError, RuntimeError, OSError) as exc:
        LOG.error("failed_to_cache_session", name=session_name, error=str(exc))
        raise


def load_session(name: str) -> SessionLike:
    """Load a cached session.

    Storage location depends on GRAFTPUNK_STORAGE_BACKEND:
    - "local" (default): Local filesystem
    - "supabase": Supabase Storage (file-pair pattern)
    - "s3": S3-compatible storage (file-pair pattern)

    Security Notes:
        This function deserializes pickle data. Pickle can execute arbitrary
        code during deserialization.

        Threat Model:
        - Sessions are encrypted with Fernet (AES-128-CBC + HMAC)
        - SHA256 checksum validation before unpickling (defense-in-depth)
        - Runtime validation is performed after unpickling to detect corrupted data
        - Fernet provides HMAC authentication (SHA256) to detect tampering

        Recommendation: Only run this tool on trusted machines.

    Args:
        name: Session name.

    Returns:
        Loaded session object.

    Raises:
        SessionNotFoundError: If session file doesn't exist.
        SessionExpiredError: If session cannot be decrypted or has invalid structure.
    """
    backend = _get_session_storage_backend()
    settings = get_settings()

    try:
        # Load encrypted data from backend
        encrypted_data, metadata = backend.load_session(name)

        # Decrypt
        try:
            decrypted_data = decrypt_data(encrypted_data)
        except EncryptionError as exc:
            LOG.error("session_decryption_failed", name=name, backend=settings.storage_backend)
            raise SessionExpiredError(
                f"Session '{name}' cannot be decrypted. The session file may be corrupted "
                "or the encryption key has changed. Please run 'graftpunk clear' and re-login."
            ) from exc

        # Verify checksum (required for new sessions, skipped for legacy)
        # Legacy sessions (loaded from flat file structure) have empty checksum
        if metadata.checksum:
            actual_checksum = hashlib.sha256(decrypted_data).hexdigest()
            if actual_checksum != metadata.checksum:
                LOG.error("session_checksum_mismatch", name=name, backend=settings.storage_backend)
                raise SessionExpiredError(
                    f"Session '{name}' failed integrity check. Run 'gp clear' and re-login."
                )
        else:
            # Legacy session without checksum - log warning but allow loading
            LOG.warning(
                "session_checksum_missing_legacy",
                name=name,
                hint="Consider re-saving session to add checksum",
            )

        # Unpickle (encrypted data has already been validated via Fernet MAC)
        session = pickle.loads(decrypted_data)  # noqa: S301

        # Runtime validation: verify unpickled object has expected attributes
        if not hasattr(session, "cookies") or not hasattr(session, "headers"):
            raise SessionExpiredError(
                f"Session '{name}' has invalid structure. Run 'gp clear' and re-login."
            )

        LOG.info("successfully_loaded_session", name=name, backend=settings.storage_backend)
        return session

    except (SessionNotFoundError, SessionExpiredError):
        raise
    except (pickle.UnpicklingError, RuntimeError) as exc:
        LOG.error(
            "failed_to_load_session",
            name=name,
            error=str(exc),
            backend=settings.storage_backend,
        )
        raise SessionExpiredError(f"Failed to load session '{name}': {exc}") from exc


def load_session_for_api(name: str) -> requests.Session:
    """Load cached session for API use (no browser required).

    This extracts cookies and headers from a cached BrowserSession
    and creates a GraftpunkSession that can be used for API calls
    without launching a browser. If the cached session has header
    profiles (captured during login), they are applied automatically.

    Args:
        name: Session name (without .session.pickle extension).

    Returns:
        GraftpunkSession with cookies, headers, and header profiles
        from the cached session.

    Raises:
        SessionNotFoundError: If session file doesn't exist.
        SessionExpiredError: If session cannot be unpickled.
    """
    try:
        browser_session = load_session(name)
    except FileNotFoundError as exc:
        LOG.error("session_not_found_for_api", name=name)
        raise SessionNotFoundError(
            f"No cached session found for '{name}'. Please login first."
        ) from exc

    # Extract header profiles if present (from sessions created after header capture feature)
    header_profiles = getattr(browser_session, "_gp_header_profiles", {})

    # Create GraftpunkSession with header profiles for auto-detection
    from graftpunk.graftpunk_session import GraftpunkSession

    api_session = GraftpunkSession(header_profiles=header_profiles)

    # Copy cookies from browser session
    if hasattr(browser_session, "cookies"):
        api_session.cookies = browser_session.cookies
        LOG.debug(
            "copied_cookies_from_session",
            cookie_count=len(browser_session.cookies),
        )

    # Copy headers from browser session, but skip requests-library defaults
    # that would clobber browser identity headers extracted from profiles.
    # The pickled BrowserSession (a requests.Session) carries default headers
    # like User-Agent: python-requests/2.x — copying them overwrites the
    # Chrome UA that _apply_browser_identity() set during GraftpunkSession init.
    if hasattr(browser_session, "headers"):
        _requests_defaults = requests.utils.default_headers()
        for key, value in browser_session.headers.items():
            if key in _requests_defaults and _requests_defaults[key] == value:
                continue
            api_session.headers[key] = value
        LOG.debug("copied_headers_from_session")

    # Copy cached tokens and CSRF tokens from browser session
    from graftpunk.tokens import _CACHE_ATTR, _CSRF_TOKENS_ATTR

    token_cache = getattr(browser_session, _CACHE_ATTR, None)
    if token_cache:
        setattr(api_session, _CACHE_ATTR, token_cache)
        LOG.debug("copied_cached_tokens_from_session", count=len(token_cache))

    csrf_tokens = getattr(browser_session, _CSRF_TOKENS_ATTR, None)
    if csrf_tokens is not None:
        setattr(api_session, _CSRF_TOKENS_ATTR, dict(csrf_tokens))
        LOG.debug("copied_csrf_tokens_from_session", count=len(csrf_tokens))

    LOG.info(
        "created_api_session_from_cached_session",
        name=name,
        has_header_profiles=bool(header_profiles),
        profile_count=len(header_profiles),
    )
    return api_session


def update_session_cookies(api_session: requests.Session, session_name: str) -> None:
    """Persist an API session's cookies and token cache back to the session cache.

    Loads the original cached session (preserving browser metadata),
    updates its cookies and token cache from the API session, and saves it back.
    This is best-effort — failures are logged but do not raise.

    Args:
        api_session: The API session with potentially updated cookies and token cache.
        session_name: Name of the cached session to update.
    """
    try:
        original = load_session(session_name)
    except Exception as exc:  # noqa: BLE001 — best-effort save
        LOG.warning(
            "session_save_skipped_load_failed",
            session_name=session_name,
            error=str(exc),
        )
        return

    try:
        from graftpunk.tokens import _CACHE_ATTR, _CSRF_TOKENS_ATTR

        original.cookies.update(api_session.cookies)
        # Persist token cache and CSRF tokens from working session
        token_cache = getattr(api_session, _CACHE_ATTR, None)
        if token_cache is not None:
            setattr(original, _CACHE_ATTR, token_cache)
        csrf_tokens = getattr(api_session, _CSRF_TOKENS_ATTR, None)
        if csrf_tokens is not None:
            setattr(original, _CSRF_TOKENS_ATTR, dict(csrf_tokens))
        cache_session(original, session_name)
        LOG.info("session_cookies_updated", session_name=session_name)
    except Exception as exc:  # noqa: BLE001 — best-effort save
        LOG.warning(
            "session_save_failed",
            session_name=session_name,
            error=str(exc),
        )


def list_sessions() -> list[str]:
    """List all cached session names.

    Returns:
        Sorted list of session names.
    """
    backend = _get_session_storage_backend()
    return backend.list_sessions()


def list_sessions_with_metadata() -> list[dict[str, Any]]:
    """List all cached sessions with metadata.

    Returns:
        List of dicts with session metadata including:
        - name, domain, current_url
        - cookie_count, cookie_domains
        - created_at, modified_at
        - path (to session directory or file)
    """
    backend = _get_session_storage_backend()
    settings = get_settings()
    names = backend.list_sessions()

    results = []
    for name in names:
        metadata = backend.get_session_metadata(name)
        if metadata is not None:
            results.append(
                {
                    "name": metadata.name,
                    "checksum": metadata.checksum,
                    "created_at": metadata.created_at.isoformat(),
                    "modified_at": metadata.modified_at.isoformat(),
                    "expires_at": metadata.expires_at.isoformat() if metadata.expires_at else None,
                    "domain": metadata.domain,
                    "current_url": metadata.current_url,
                    "cookie_count": metadata.cookie_count,
                    "cookie_domains": metadata.cookie_domains,
                    "status": metadata.status,
                    "path": str(settings.sessions_dir / name),
                }
            )

    return sorted(results, key=lambda x: x.get("modified_at", ""), reverse=True)


def clear_session_cache(session_name: str | None = None) -> list[str]:
    """Clear cached sessions.

    Args:
        session_name: If provided, clear only this session. Otherwise, clear all.

    Returns:
        List of removed session names.
    """
    backend = _get_session_storage_backend()
    removed: list[str] = []

    if session_name:
        # Clear specific session
        if backend.delete_session(session_name):
            removed.append(session_name)
        return removed

    # Clear all sessions
    for name in backend.list_sessions():
        if backend.delete_session(name):
            removed.append(name)

    return removed


def update_session_status(name: str, status: str) -> None:
    """Update session status in metadata.

    Args:
        name: Session name.
        status: New status ("active" or "logged_out").

    Raises:
        SessionNotFoundError: If session doesn't exist.
        ValueError: If status is not a valid value.
    """
    if status not in ("active", "logged_out"):
        raise ValueError(f"Invalid status '{status}'. Must be 'active' or 'logged_out'")

    backend = _get_session_storage_backend()
    if not backend.update_session_metadata(name, status=status):
        raise SessionNotFoundError(f"Session '{name}' not found")

    LOG.info("updated_session_status", name=name, status=status)



"""Tests for local storage backend."""

import json
from datetime import UTC, datetime, timedelta

import pytest

from graftpunk.exceptions import SessionExpiredError, SessionNotFoundError
from graftpunk.storage.base import SessionMetadata
from graftpunk.storage.local import LocalSessionStorage


class TestLocalSessionStorage:
    """Tests for LocalSessionStorage."""

    @pytest.fixture
    def storage(self, tmp_path):
        """Create a LocalSessionStorage instance."""
        return LocalSessionStorage(base_dir=tmp_path)

    @pytest.fixture
    def sample_metadata(self):
        """Create sample session metadata."""
        now = datetime.now(UTC)
        return SessionMetadata(
            name="test-session",
            checksum="abc123",
            created_at=now,
            modified_at=now,
            expires_at=now + timedelta(hours=24),
            domain="example.com",
            current_url="https://example.com/dashboard",
            cookie_count=5,
            cookie_domains=["example.com", ".example.com"],
            status="active",
        )

    def test_save_and_load_session(self, storage, sample_metadata):
        """Test saving and loading a session."""
        encrypted_data = b"encrypted session data"

        # Save
        location = storage.save_session(
            "test-session",
            encrypted_data,
            sample_metadata,
        )
        assert "test-session" in location

        # Load
        loaded_data, loaded_metadata = storage.load_session("test-session")
        assert loaded_data == encrypted_data
        assert loaded_metadata.name == sample_metadata.name
        assert loaded_metadata.checksum == sample_metadata.checksum
        assert loaded_metadata.domain == sample_metadata.domain

    def test_list_sessions(self, storage, sample_metadata):
        """Test listing sessions."""
        # Initially empty
        assert storage.list_sessions() == []

        # Save some sessions
        storage.save_session("session-a", b"data", sample_metadata)
        storage.save_session("session-b", b"data", sample_metadata)

        # List should return sorted names
        sessions = storage.list_sessions()
        assert sessions == ["session-a", "session-b"]

    def test_delete_session(self, storage, sample_metadata):
        """Test deleting a session."""
        storage.save_session("to-delete", b"data", sample_metadata)
        assert "to-delete" in storage.list_sessions()

        # Delete
        result = storage.delete_session("to-delete")
        assert result is True
        assert "to-delete" not in storage.list_sessions()

        # Delete non-existent
        result = storage.delete_session("non-existent")
        assert result is False

    def test_get_session_metadata(self, storage, sample_metadata):
        """Test getting session metadata without full load."""
        storage.save_session("test-session", b"data", sample_metadata)

        metadata = storage.get_session_metadata("test-session")
        assert metadata is not None
        assert metadata.name == "test-session"
        assert metadata.domain == sample_metadata.domain
        assert metadata.cookie_count == sample_metadata.cookie_count

        # Non-existent session
        assert storage.get_session_metadata("non-existent") is None

    def test_update_session_metadata(self, storage, sample_metadata):
        """Test updating session status."""
        storage.save_session("update-test", b"data", sample_metadata)

        # Update status
        result = storage.update_session_metadata("update-test", status="logged_out")
        assert result is True

        # Verify update
        metadata = storage.get_session_metadata("update-test")
        assert metadata is not None
        assert metadata.status == "logged_out"

    def test_update_invalid_status_raises_error(self, storage, sample_metadata):
        """Test that invalid status raises ValueError."""
        storage.save_session("invalid-status", b"data", sample_metadata)

        with pytest.raises(ValueError, match="Invalid status"):
            storage.update_session_metadata("invalid-status", status="invalid")

    def test_load_nonexistent_session_raises_error(self, storage):
        """Test that loading non-existent session raises SessionNotFoundError."""
        with pytest.raises(SessionNotFoundError):
            storage.load_session("non-existent")

    def test_load_expired_session_raises_error(self, storage):
        """Test that loading expired session raises SessionExpiredError."""
        now = datetime.now(UTC)
        expired_metadata = SessionMetadata(
            name="expired",
            checksum="abc",
            created_at=now - timedelta(hours=48),
            modified_at=now - timedelta(hours=48),
            expires_at=now - timedelta(hours=24),  # Expired
            domain="example.com",
            current_url=None,
            cookie_count=0,
            cookie_domains=[],
            status="active",
        )

        storage.save_session("expired", b"data", expired_metadata)

        with pytest.raises(SessionExpiredError):
            storage.load_session("expired")


class TestLegacyFlatFile:
    """Tests for legacy flat file session loading and management."""

    @pytest.fixture
    def storage(self, tmp_path):
        """Create a LocalSessionStorage instance."""
        return LocalSessionStorage(base_dir=tmp_path)

    def test_load_legacy_session(self, storage, tmp_path):
        """Test loading a session from legacy flat file structure."""
        # Create a legacy flat file
        legacy_path = tmp_path / "my-provider.session.pickle"
        legacy_path.write_bytes(b"legacy-encrypted-data")

        data, metadata = storage.load_session("my-provider")
        assert data == b"legacy-encrypted-data"
        assert metadata.name == "my-provider"
        assert metadata.checksum == ""
        assert metadata.expires_at is None
        assert metadata.status == "active"

    def test_list_sessions_includes_legacy_files(self, storage, tmp_path):
        """Test that list_sessions includes legacy flat file sessions."""
        # Create a legacy flat file
        legacy_path = tmp_path / "legacy-provider.session.pickle"
        legacy_path.write_bytes(b"data")

        # Create a modern directory session
        modern_dir = tmp_path / "modern-provider"
        modern_dir.mkdir()
        (modern_dir / "session.pickle").write_bytes(b"data")

        sessions = storage.list_sessions()
        assert "legacy-provider" in sessions
        assert "modern-provider" in sessions
        assert sessions == sorted(sessions)

    def test_delete_legacy_session(self, storage, tmp_path):
        """Test deleting a legacy flat file session."""
        legacy_path = tmp_path / "old-session.session.pickle"
        legacy_path.write_bytes(b"data")

        result = storage.delete_session("old-session")
        assert result is True
        assert not legacy_path.exists()

    def test_delete_legacy_session_os_error(self, storage, tmp_path):
        """Test that OSError during legacy file deletion returns False."""
        # Session does not exist as directory or legacy file
        result = storage.delete_session("totally-nonexistent")
        assert result is False


class TestLocalEdgeCases:
    """Tests for edge cases and error handling in local storage."""

    @pytest.fixture
    def storage(self, tmp_path):
        """Create a LocalSessionStorage instance."""
        return LocalSessionStorage(base_dir=tmp_path)

    @pytest.fixture
    def sample_metadata(self):
        """Create sample session metadata."""
        now = datetime.now(UTC)
        return SessionMetadata(
            name="test-session",
            checksum="abc123",
            created_at=now,
            modified_at=now,
            expires_at=now + timedelta(hours=24),
            domain="example.com",
            current_url="https://example.com/dashboard",
            cookie_count=5,
            cookie_domains=["example.com", ".example.com"],
            status="active",
        )

    def test_load_session_missing_metadata_file(self, storage, tmp_path):
        """Test load raises SessionExpiredError when metadata.json is missing."""
        session_dir = tmp_path / "no-meta"
        session_dir.mkdir()
        (session_dir / "session.pickle").write_bytes(b"data")
        # No metadata.json

        with pytest.raises(SessionExpiredError, match="missing metadata"):
            storage.load_session("no-meta")

    def test_load_session_invalid_metadata_json(self, storage, tmp_path):
        """Test load raises SessionExpiredError when metadata.json is invalid JSON."""
        session_dir = tmp_path / "bad-meta"
        session_dir.mkdir()
        (session_dir / "session.pickle").write_bytes(b"data")
        (session_dir / "metadata.json").write_text("not valid json {{{")

        with pytest.raises(SessionExpiredError, match="invalid metadata"):
            storage.load_session("bad-meta")

    def test_load_session_pickle_missing_in_dir(self, storage, tmp_path):
        """Test load raises SessionNotFoundError when dir exists but pickle is missing."""
        session_dir = tmp_path / "empty-dir"
        session_dir.mkdir()
        # No session.pickle

        with pytest.raises(SessionNotFoundError, match="not found"):
            storage.load_session("empty-dir")

    def test_load_session_invalid_expires_at_continues(self, storage, tmp_path):
        """Test load continues when expires_at is an unparseable string."""
        session_dir = tmp_path / "bad-expiry"
        session_dir.mkdir()
        (session_dir / "session.pickle").write_bytes(b"data")
        metadata_dict = {
            "name": "bad-expiry",
            "checksum": "abc",
            "created_at": datetime.now(UTC).isoformat(),
            "modified_at": datetime.now(UTC).isoformat(),
            "expires_at": "not-a-date",
            "domain": None,
            "current_url": None,
            "cookie_count": 0,
            "cookie_domains": [],
            "status": "active",
        }
        with (session_dir / "metadata.json").open("w") as f:
            json.dump(metadata_dict, f)

        data, metadata = storage.load_session("bad-expiry")
        assert data == b"data"
        assert metadata.name == "bad-expiry"

    def test_get_session_metadata_json_decode_error(self, storage, tmp_path):
        """Test get_session_metadata returns None on JSON decode error."""
        session_dir = tmp_path / "corrupt"
        session_dir.mkdir()
        (session_dir / "metadata.json").write_text("{corrupt json")

        result = storage.get_session_metadata("corrupt")
        assert result is None

    def test_update_session_metadata_nonexistent_returns_false(self, storage):
        """Test update_session_metadata returns False for nonexistent session."""
        result = storage.update_session_metadata("nonexistent", status="active")
        assert result is False

    def test_update_session_metadata_corrupt_json_returns_false(self, storage, tmp_path):
        """Test update_session_metadata returns False on corrupt metadata.json."""
        session_dir = tmp_path / "corrupt-update"
        session_dir.mkdir()
        (session_dir / "metadata.json").write_text("{not valid json")

        result = storage.update_session_metadata("corrupt-update", status="active")
        assert result is False

    def test_update_session_metadata_no_status_updates_modified_at(
        self, storage, sample_metadata, tmp_path
    ):
        """Test update_session_metadata with no status still updates modified_at."""
        storage.save_session("just-touch", b"data", sample_metadata)

        # Read original modified_at
        metadata_before = storage.get_session_metadata("just-touch")
        assert metadata_before is not None

        result = storage.update_session_metadata("just-touch")
        assert result is True

        metadata_after = storage.get_session_metadata("just-touch")
        assert metadata_after is not None
        # modified_at should be updated
        assert metadata_after.modified_at >= metadata_before.modified_at

    def test_list_sessions_base_dir_not_exists(self, tmp_path):
        """Test list_sessions returns empty when base_dir doesn't exist."""
        import shutil

        storage = LocalSessionStorage(base_dir=tmp_path / "sessions")
        # Remove the directory that was created during init
        shutil.rmtree(tmp_path / "sessions")

        sessions = storage.list_sessions()
        assert sessions == []

    def test_load_session_no_expiry_succeeds(self, storage, tmp_path):
        """Test load succeeds when expires_at is None (no TTL)."""
        now = datetime.now(UTC)
        no_expiry_metadata = SessionMetadata(
            name="no-expiry",
            checksum="abc",
            created_at=now,
            modified_at=now,
            expires_at=None,
            domain="example.com",
            current_url=None,
            cookie_count=0,
            cookie_domains=[],
            status="active",
        )

        storage.save_session("no-expiry", b"session-data", no_expiry_metadata)

        data, metadata = storage.load_session("no-expiry")
        assert data == b"session-data"
        assert metadata.expires_at is None


class TestSessionMetadata:
    """Tests for SessionMetadata dataclass."""

    def test_metadata_is_frozen(self):
        """Test that SessionMetadata is immutable."""
        now = datetime.now(UTC)
        metadata = SessionMetadata(
            name="test",
            checksum="abc",
            created_at=now,
            modified_at=now,
            expires_at=None,
            domain="example.com",
            current_url=None,
            cookie_count=0,
            cookie_domains=[],
        )

        with pytest.raises(AttributeError):
            metadata.name = "changed"  # type: ignore

    def test_metadata_defaults(self):
        """Test SessionMetadata default values."""
        now = datetime.now(UTC)
        metadata = SessionMetadata(
            name="test",
            checksum="abc",
            created_at=now,
            modified_at=now,
            expires_at=None,
            domain=None,
            current_url=None,
            cookie_count=0,
            cookie_domains=[],
        )

        assert metadata.status == "active"



"""Tests for CSRF token and dynamic header extraction (non-browser)."""

from __future__ import annotations

import time
from unittest.mock import MagicMock, patch

import pytest
import requests

from graftpunk.tokens import (
    CachedToken,
    Token,
    TokenConfig,
    _store_csrf_token,
    clear_cached_tokens,
    extract_token,
    prepare_session,
)


class TestToken:
    """Tests for Token dataclass validation and factory methods."""

    def test_token_validation_empty_name(self) -> None:
        with pytest.raises(ValueError, match="Token.name must be non-empty"):
            Token(name="", source="cookie", cookie_name="csrf")

    def test_token_validation_invalid_source(self) -> None:
        with pytest.raises(ValueError, match="Token.source must be"):
            Token(name="X-CSRF", source="invalid")

    def test_token_page_requires_pattern(self) -> None:
        with pytest.raises(ValueError, match="requires a pattern"):
            Token(name="X-CSRF", source="page")

    def test_token_cookie_requires_cookie_name(self) -> None:
        with pytest.raises(ValueError, match="requires cookie_name"):
            Token(name="X-CSRF", source="cookie")

    def test_token_response_header_requires_header(self) -> None:
        with pytest.raises(ValueError, match="requires response_header"):
            Token(name="X-CSRF", source="response_header")

    def test_from_meta_tag(self) -> None:
        token = Token.from_meta_tag("csrf-token", "X-CSRF-Token")
        assert token.name == "X-CSRF-Token"
        assert token.source == "page"
        assert token.pattern is not None
        assert "csrf\\-token" in token.pattern
        assert token.page_url == "/"
        assert token.cache_duration == 300

    def test_from_cookie(self) -> None:
        token = Token.from_cookie("csrftoken", "X-CSRFToken")
        assert token.name == "X-CSRFToken"
        assert token.source == "cookie"
        assert token.cookie_name == "csrftoken"
        assert token.cache_duration == 300

    def test_from_js_variable(self) -> None:
        token = Token.from_js_variable(r'var csrf = "([^"]+)"', "X-CSRF", page_url="/app")
        assert token.name == "X-CSRF"
        assert token.source == "page"
        assert token.pattern == r'var csrf = "([^"]+)"'
        assert token.page_url == "/app"

    def test_from_response_header(self) -> None:
        token = Token.from_response_header("X-CSRF-Token", "X-Request-CSRF")
        assert token.name == "X-Request-CSRF"
        assert token.source == "response_header"
        assert token.response_header == "X-CSRF-Token"

    def test_valid_page_token(self) -> None:
        token = Token(name="X-CSRF", source="page", pattern=r"csrf=(\w+)")
        assert token.name == "X-CSRF"
        assert token.source == "page"

    def test_valid_cookie_token(self) -> None:
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrftoken")
        assert token.name == "X-CSRF"
        assert token.cookie_name == "csrftoken"

    def test_extraction_defaults_to_auto(self) -> None:
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrf")
        assert token.extraction == "auto"

    def test_extraction_accepts_http(self) -> None:
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrf", extraction="http")
        assert token.extraction == "http"

    def test_extraction_accepts_browser(self) -> None:
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrf", extraction="browser")
        assert token.extraction == "browser"

    def test_extraction_invalid_value_raises(self) -> None:
        with pytest.raises(ValueError, match="Token.extraction must be"):
            Token(name="X-CSRF", source="cookie", cookie_name="csrf", extraction="invalid")

    def test_from_js_variable_with_browser_extraction(self) -> None:
        token = Token.from_js_variable(r'csrf = "([^"]+)"', "X-CSRF", extraction="browser")
        assert token.extraction == "browser"

    def test_from_meta_tag_with_extraction(self) -> None:
        token = Token.from_meta_tag("csrf-token", "X-CSRF", extraction="browser")
        assert token.extraction == "browser"

    def test_from_response_header_with_extraction(self) -> None:
        token = Token.from_response_header("X-Token", "X-Req", extraction="http")
        assert token.extraction == "http"

    def test_cache_duration_must_be_positive(self) -> None:
        with pytest.raises(ValueError, match="cache_duration must be positive"):
            Token(
                name="X-CSRF",
                source="page",
                pattern=r"csrf=([^&]+)",
                cache_duration=0,
            )

    def test_negative_cache_duration_rejected(self) -> None:
        with pytest.raises(ValueError, match="cache_duration must be positive"):
            Token(
                name="X-CSRF",
                source="page",
                pattern=r"csrf=([^&]+)",
                cache_duration=-1,
            )


class TestTokenConfig:
    """Tests for TokenConfig validation."""

    def test_empty_tokens_raises(self) -> None:
        with pytest.raises(ValueError, match="TokenConfig.tokens must be non-empty"):
            TokenConfig(tokens=())

    def test_valid_config(self) -> None:
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrf")
        config = TokenConfig(tokens=(token,))
        assert len(config.tokens) == 1
        assert config.tokens[0].name == "X-CSRF"


class TestCachedToken:
    """Tests for CachedToken TTL logic."""

    def test_not_expired(self) -> None:
        cached = CachedToken(name="X-CSRF", value="abc123", extracted_at=time.time(), ttl=300)
        assert not cached.is_expired

    def test_expired(self) -> None:
        cached = CachedToken(name="X-CSRF", value="abc123", extracted_at=time.time() - 400, ttl=300)
        assert cached.is_expired


class TestExtractToken:
    """Tests for extract_token function."""

    def test_extract_from_cookie(self) -> None:
        session = requests.Session()
        session.cookies.set("csrftoken", "secret123")
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrftoken")
        value = extract_token(session, token, "https://example.com")
        assert value == "secret123"

    def test_extract_from_cookie_missing(self) -> None:
        session = requests.Session()
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrftoken")
        with pytest.raises(ValueError, match="Cookie 'csrftoken' not found"):
            extract_token(session, token, "https://example.com")

    def test_extract_from_page(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf_token = "([^"]+)"',
            page_url="/dashboard",
        )
        mock_resp = MagicMock()
        mock_resp.text = 'var csrf_token = "abc123xyz";'
        mock_resp.raise_for_status = MagicMock()

        with patch.object(session, "get", return_value=mock_resp) as mock_get:
            value = extract_token(session, token, "https://example.com")
            assert value == "abc123xyz"
            mock_get.assert_called_once_with("https://example.com/dashboard", timeout=15)

    def test_extract_from_page_pattern_not_found(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf_token = "([^"]+)"',
            page_url="/dashboard",
            extraction="http",
        )
        mock_resp = MagicMock()
        mock_resp.text = "<html>no token here</html>"
        mock_resp.raise_for_status = MagicMock()

        with (
            patch.object(session, "get", return_value=mock_resp),
            pytest.raises(ValueError, match="Token pattern not found"),
        ):
            extract_token(session, token, "https://example.com")

    def test_extract_from_response_header(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="response_header",
            response_header="X-CSRF-Token",
            page_url="/api",
        )
        mock_resp = MagicMock()
        mock_resp.headers = {"X-CSRF-Token": "header_value_123"}

        with patch.object(session, "head", return_value=mock_resp) as mock_head:
            value = extract_token(session, token, "https://example.com")
            assert value == "header_value_123"
            mock_head.assert_called_once_with(
                "https://example.com/api", timeout=10, allow_redirects=True
            )

    def test_extract_from_response_header_missing(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="response_header",
            response_header="X-CSRF-Token",
            page_url="/api",
        )
        mock_resp = MagicMock()
        mock_resp.headers = {}

        with (
            patch.object(session, "head", return_value=mock_resp),
            pytest.raises(ValueError, match="Header 'X-CSRF-Token' not found"),
        ):
            extract_token(session, token, "https://example.com")

    def test_extract_unknown_source(self) -> None:
        """Test that an unknown source raises ValueError.

        We bypass __post_init__ validation using object.__setattr__ on a frozen
        dataclass to test the extract_token function's own guard clause.
        """
        token = Token(name="X-CSRF", source="cookie", cookie_name="x")
        # Force an invalid source past validation
        object.__setattr__(token, "source", "unknown")
        session = requests.Session()
        with pytest.raises(ValueError, match="Unknown token source"):
            extract_token(session, token, "https://example.com")


class TestPrepareSession:
    """Tests for prepare_session function."""

    def test_extracts_and_stores_token(self) -> None:
        session = requests.Session()
        session.cookies.set("csrftoken", "injected_value")
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrftoken")
        config = TokenConfig(tokens=(token,))

        result = prepare_session(session, config, "https://example.com")
        assert result is session
        csrf_tokens = getattr(session, "_gp_csrf_tokens", {})
        assert csrf_tokens["X-CSRF"] == "injected_value"
        # Token should NOT be in session.headers (method-scoped injection only)
        assert "X-CSRF" not in session.headers

    def test_uses_cached_token(self) -> None:
        session = requests.Session()
        # Pre-populate cache
        cached = CachedToken(name="X-CSRF", value="cached_value", extracted_at=time.time(), ttl=300)
        session._gp_cached_tokens = {"X-CSRF": cached}  # type: ignore[attr-defined]

        token = Token(name="X-CSRF", source="cookie", cookie_name="csrftoken")
        config = TokenConfig(tokens=(token,))

        # Should use cache, not try to extract (no cookie set, would fail otherwise)
        prepare_session(session, config, "https://example.com")
        csrf_tokens = getattr(session, "_gp_csrf_tokens", {})
        assert csrf_tokens["X-CSRF"] == "cached_value"

    def test_injects_expired_cached_token_without_extraction(self) -> None:
        """Expired cached token is injected without triggering extraction (EAFP)."""
        session = requests.Session()
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrf")
        config = TokenConfig(tokens=(token,))

        # Pre-populate with expired cache
        from graftpunk.tokens import _CACHE_ATTR

        expired_cache = {
            "X-CSRF": CachedToken(name="X-CSRF", value="old_value", extracted_at=0, ttl=1)
        }
        setattr(session, _CACHE_ATTR, expired_cache)

        # prepare_session should inject the expired value without calling extract_token
        with patch("graftpunk.tokens.extract_token") as mock_extract:
            prepare_session(session, config, "https://example.com")

        mock_extract.assert_not_called()
        csrf_tokens = getattr(session, "_gp_csrf_tokens", {})
        assert csrf_tokens["X-CSRF"] == "old_value"

    def test_no_cache_still_extracts(self) -> None:
        """Token with no cache at all still triggers extraction."""
        session = requests.Session()
        session.cookies.set("csrf", "fresh_value")
        token = Token(name="X-CSRF", source="cookie", cookie_name="csrf")
        config = TokenConfig(tokens=(token,))

        prepare_session(session, config, "https://example.com")
        csrf_tokens = getattr(session, "_gp_csrf_tokens", {})
        assert csrf_tokens["X-CSRF"] == "fresh_value"

    def test_multiple_tokens(self) -> None:
        session = requests.Session()
        session.cookies.set("csrftoken", "csrf_val")
        session.cookies.set("session_id", "sess_val")

        token1 = Token(name="X-CSRF", source="cookie", cookie_name="csrftoken")
        token2 = Token(name="X-Session", source="cookie", cookie_name="session_id")
        config = TokenConfig(tokens=(token1, token2))

        prepare_session(session, config, "https://example.com")
        csrf_tokens = getattr(session, "_gp_csrf_tokens", {})
        assert csrf_tokens["X-CSRF"] == "csrf_val"
        assert csrf_tokens["X-Session"] == "sess_val"


class TestClearCachedTokens:
    """Tests for clear_cached_tokens function."""

    def test_clears_cache(self) -> None:
        session = requests.Session()
        cached = CachedToken(name="X-CSRF", value="val", extracted_at=time.time(), ttl=300)
        session._gp_cached_tokens = {"X-CSRF": cached}  # type: ignore[attr-defined]

        clear_cached_tokens(session)
        assert session._gp_cached_tokens == {}  # type: ignore[attr-defined]

    def test_clears_csrf_tokens(self) -> None:
        session = requests.Session()
        session._gp_csrf_tokens = {"X-CSRF": "secret"}  # type: ignore[attr-defined]

        clear_cached_tokens(session)
        assert session._gp_csrf_tokens == {}  # type: ignore[attr-defined]

    def test_clears_both_cache_and_csrf_tokens(self) -> None:
        session = requests.Session()
        cached = CachedToken(name="X-CSRF", value="val", extracted_at=time.time(), ttl=300)
        session._gp_cached_tokens = {"X-CSRF": cached}  # type: ignore[attr-defined]
        session._gp_csrf_tokens = {"X-CSRF": "secret"}  # type: ignore[attr-defined]

        clear_cached_tokens(session)
        assert session._gp_cached_tokens == {}  # type: ignore[attr-defined]
        assert session._gp_csrf_tokens == {}  # type: ignore[attr-defined]

    def test_no_cache_no_error(self) -> None:
        session = requests.Session()
        # Should not raise
        clear_cached_tokens(session)


class TestStoreCsrfToken:
    """Tests for _store_csrf_token function."""

    def test_creates_attribute_on_first_call(self) -> None:
        session = requests.Session()
        assert not hasattr(session, "_gp_csrf_tokens")
        _store_csrf_token(session, "X-CSRF", "secret")
        assert session._gp_csrf_tokens == {"X-CSRF": "secret"}  # type: ignore[attr-defined]

    def test_overwrites_existing_key(self) -> None:
        session = requests.Session()
        _store_csrf_token(session, "X-CSRF", "old")
        _store_csrf_token(session, "X-CSRF", "new")
        assert session._gp_csrf_tokens == {"X-CSRF": "new"}  # type: ignore[attr-defined]

    def test_adds_new_key(self) -> None:
        session = requests.Session()
        _store_csrf_token(session, "X-CSRF", "val1")
        _store_csrf_token(session, "X-Token", "val2")
        assert session._gp_csrf_tokens == {"X-CSRF": "val1", "X-Token": "val2"}  # type: ignore[attr-defined]


class TestExtractTokenExtractionField:
    """Tests for extract_token respecting the extraction field."""

    def test_extraction_http_uses_session_get(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="http",
        )
        mock_resp = MagicMock()
        mock_resp.text = 'csrf = "abc";'
        mock_resp.raise_for_status = MagicMock()

        with patch.object(session, "get", return_value=mock_resp):
            value = extract_token(session, token, "https://example.com")
        assert value == "abc"

    def test_extraction_browser_raises_needs_browser(self) -> None:
        from graftpunk.tokens import _BrowserExtractionNeeded as BrowserExtractionNeeded

        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="browser",
        )
        with pytest.raises(BrowserExtractionNeeded):
            extract_token(session, token, "https://example.com")

    def test_extraction_auto_tries_http_first(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="auto",
        )
        mock_resp = MagicMock()
        mock_resp.text = 'csrf = "abc";'
        mock_resp.raise_for_status = MagicMock()

        with patch.object(session, "get", return_value=mock_resp):
            value = extract_token(session, token, "https://example.com")
        assert value == "abc"

    def test_extraction_auto_raises_needs_browser_on_timeout(self) -> None:
        from graftpunk.tokens import _BrowserExtractionNeeded as BrowserExtractionNeeded

        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="auto",
        )

        with (
            patch.object(session, "get", side_effect=requests.exceptions.ReadTimeout("timeout")),
            pytest.raises(BrowserExtractionNeeded),
        ):
            extract_token(session, token, "https://example.com")

    def test_extraction_auto_raises_needs_browser_on_403(self) -> None:
        from graftpunk.tokens import _BrowserExtractionNeeded as BrowserExtractionNeeded

        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="auto",
        )
        mock_resp = MagicMock()
        mock_resp.status_code = 403
        mock_resp.raise_for_status = MagicMock(
            side_effect=requests.exceptions.HTTPError(response=mock_resp)
        )

        with (
            patch.object(session, "get", return_value=mock_resp),
            pytest.raises(BrowserExtractionNeeded),
        ):
            extract_token(session, token, "https://example.com")

    def test_extraction_http_raises_original_on_timeout(self) -> None:
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="http",
        )

        with (
            patch.object(session, "get", side_effect=requests.exceptions.ReadTimeout("timeout")),
            pytest.raises(requests.exceptions.ReadTimeout),
        ):
            extract_token(session, token, "https://example.com")

    def test_cookie_source_ignores_extraction_field(self) -> None:
        session = requests.Session()
        session.cookies.set("csrf", "val")
        token = Token(
            name="X-CSRF",
            source="cookie",
            cookie_name="csrf",
            extraction="browser",
        )
        value = extract_token(session, token, "https://example.com")
        assert value == "val"

    def test_extraction_auto_response_header_falls_back_on_error(self) -> None:
        from graftpunk.tokens import _BrowserExtractionNeeded as BrowserExtractionNeeded

        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="response_header",
            response_header="X-Token",
            page_url="/api",
            extraction="auto",
        )

        with (
            patch.object(
                session,
                "head",
                side_effect=requests.exceptions.ConnectionError("refused"),
            ),
            pytest.raises(BrowserExtractionNeeded),
        ):
            extract_token(session, token, "https://example.com")

    def test_extraction_auto_falls_back_on_pattern_not_found(self) -> None:
        """extraction=auto falls back to browser when HTTP 200 but pattern not found."""
        from graftpunk.tokens import _BrowserExtractionNeeded as BrowserExtractionNeeded

        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="page",
            pattern=r'csrf = "([^"]+)"',
            page_url="/app",
            extraction="auto",
        )
        mock_resp = MagicMock()
        mock_resp.text = "<html>no csrf here</html>"
        mock_resp.raise_for_status = MagicMock()

        with (
            patch.object(session, "get", return_value=mock_resp),
            pytest.raises(BrowserExtractionNeeded),
        ):
            extract_token(session, token, "https://example.com")

    def test_extraction_http_response_header_raises_on_error(self) -> None:
        """extraction=http with source=response_header raises on network error."""
        session = requests.Session()
        token = Token(
            name="X-CSRF",
            source="response_header",
            response_header="X-Token",
            page_url="/api",
            extraction="http",
        )

        with (
            patch.object(
                session,
                "head",
                side_effect=requests.exceptions.ConnectionError("refused"),
            ),
            pytest.raises(requests.exceptions.ConnectionError),
        ):
            extract_token(session, token, "https://example.com")



"""Tests for YAML site plugin factory (yaml_plugin.py)."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest
import requests

from graftpunk.plugins.cli_plugin import (
    CommandContext,
    LoginConfig,
    LoginStep,
    PluginConfig,
    SitePlugin,
    build_plugin_config,
)
from graftpunk.plugins.yaml_loader import YAMLCommandDef, YAMLParamDef
from graftpunk.plugins.yaml_plugin import _convert_params, create_yaml_site_plugin


def _make_config(
    *,
    site_name: str = "testsite",
    session_name: str = "testsession",
    base_url: str = "https://api.example.com",
    requires_session: bool = True,
    login_config: LoginConfig | None = None,
) -> PluginConfig:
    """Helper to create a PluginConfig with sensible defaults."""
    return build_plugin_config(
        site_name=site_name,
        session_name=session_name,
        base_url=base_url,
        requires_session=requires_session,
        login_config=login_config,
    )


def _make_command(
    *,
    name: str = "test",
    url: str = "/test",
    method: str = "GET",
    params: tuple | None = None,
    headers: dict[str, str] | None = None,
    jmespath: str | None = None,
    raise_for_status: bool = True,
    timeout: float | None = None,
    max_retries: int = 0,
    rate_limit: float | None = None,
) -> YAMLCommandDef:
    """Helper to create a YAMLCommandDef."""
    return YAMLCommandDef(
        name=name,
        help_text=f"Help for {name}",
        method=method,
        url=url,
        params=params or (),
        headers=headers or {},
        jmespath=jmespath,
        raise_for_status=raise_for_status,
        timeout=timeout,
        max_retries=max_retries,
        rate_limit=rate_limit,
    )


def _mock_ctx(session: MagicMock | None = None) -> CommandContext:
    """Create a mock CommandContext wrapping a mock session."""
    if session is None:
        session = MagicMock(spec=requests.Session)
    return CommandContext(
        session=session,
        plugin_name="testsite",
        command_name="test",
        api_version=1,
    )


def _mock_response(
    *,
    json_data: object | None = None,
    text: str = "",
    headers: dict[str, str] | None = None,
    json_raises: bool = False,
) -> MagicMock:
    """Create a mock requests.Response."""
    resp = MagicMock(spec=requests.Response)
    resp.text = text
    resp.headers = headers or {}
    if json_raises:
        resp.json.side_effect = ValueError("No JSON")
    else:
        resp.json.return_value = json_data
    return resp


class TestLoginProperties:
    """Tests for login-related properties on the dynamic SitePlugin."""

    def test_login_properties_with_login_config(self) -> None:
        """When login is set, LoginConfig appears as attribute on the plugin."""
        config = _make_config(
            login_config=LoginConfig(
                steps=[
                    LoginStep(
                        fields={"username": "#user", "password": "#pass"},
                        submit="#submit",
                    )
                ],
                url="/login",
                failure="Bad login",
                success="Welcome",
            ),
        )
        plugin = create_yaml_site_plugin(config, [])

        assert plugin.login_config is not None
        assert plugin.login_config.url == "/login"
        assert len(plugin.login_config.steps) == 1
        assert plugin.login_config.steps[0].fields == {"username": "#user", "password": "#pass"}
        assert plugin.login_config.steps[0].submit == "#submit"
        assert plugin.login_config.failure == "Bad login"
        assert plugin.login_config.success == "Welcome"

    def test_login_properties_without_login_config(self) -> None:
        """When login is not configured, login_config is None."""
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [])

        assert plugin.login_config is None


class TestPluginType:
    """Tests that the factory produces proper SitePlugin instances."""

    def test_plugin_is_site_plugin_instance(self) -> None:
        """The created plugin is an instance of SitePlugin."""
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [])
        assert isinstance(plugin, SitePlugin)

    def test_plugin_has_correct_site_name(self) -> None:
        """The created plugin has the correct site_name."""
        config = _make_config(site_name="mysite")
        plugin = create_yaml_site_plugin(config, [])
        assert plugin.site_name == "mysite"

    def test_plugin_has_correct_session_name(self) -> None:
        """The created plugin has the correct session_name."""
        config = _make_config(session_name="mysession")
        plugin = create_yaml_site_plugin(config, [])
        assert plugin.session_name == "mysession"


class TestGetSession:
    """Tests for the plugin's get_session (inherited from SitePlugin)."""

    def test_get_session_with_session_name(self) -> None:
        """When requires_session is True, load_session_for_api is called."""
        config = _make_config(session_name="my_session")
        plugin = create_yaml_site_plugin(config, [])

        mock_session = MagicMock(spec=requests.Session)
        with patch(
            "graftpunk.plugins.cli_plugin.load_session_for_api",
            return_value=mock_session,
        ) as mock_load:
            result = plugin.get_session()

        mock_load.assert_called_once_with("my_session")
        assert result is mock_session

    def test_get_session_without_requires_session(self) -> None:
        """When requires_session is False, a plain requests.Session is returned."""
        config = _make_config(requires_session=False)
        plugin = create_yaml_site_plugin(config, [])

        session = plugin.get_session()
        assert isinstance(session, requests.Session)


class TestHandlerURLConstruction:
    """Tests for URL building inside the handler closure."""

    def test_base_url_prepended(self) -> None:
        """Handler prepends base_url to the command url."""
        cmd = _make_command(url="/items")
        config = _make_config(base_url="https://api.example.com")
        plugin = create_yaml_site_plugin(config, [cmd])
        commands = {c.name: c for c in plugin.get_commands()}
        handler = commands["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"ok": True})
        ctx = _mock_ctx(session)

        handler(ctx)

        call_kwargs = session.request.call_args
        assert call_kwargs.kwargs["url"] == "https://api.example.com/items"

    def test_no_base_url(self) -> None:
        """Without base_url, the command url is used as-is."""
        cmd = _make_command(url="https://other.com/data")
        config = _make_config(base_url="")
        plugin = create_yaml_site_plugin(config, [cmd])
        commands = {c.name: c for c in plugin.get_commands()}
        handler = commands["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={})
        ctx = _mock_ctx(session)

        handler(ctx)

        assert session.request.call_args.kwargs["url"] == "https://other.com/data"

    def test_url_parameter_substitution(self) -> None:
        """URL {param} placeholders are substituted with kwarg values."""
        cmd = _make_command(url="/users/{user_id}/posts/{post_id}")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data=[])
        ctx = _mock_ctx(session)

        handler(ctx, user_id=42, post_id=7)

        assert session.request.call_args.kwargs["url"] == (
            "https://api.example.com/users/42/posts/7"
        )

    def test_missing_url_parameter_raises(self) -> None:
        """Unsubstituted URL parameters raise ValueError."""
        cmd = _make_command(url="/users/{user_id}")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        ctx = _mock_ctx(session)

        with pytest.raises(ValueError, match="Missing required URL parameters: user_id"):
            handler(ctx)


class TestHandlerHeaders:
    """Tests for header expansion in the handler closure."""

    def test_plugin_and_command_headers_merged(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Plugin-level and command-level headers are merged, command wins."""
        monkeypatch.setenv("MY_TOKEN", "abc123")
        cmd = _make_command(
            url="/test",
            headers={"X-Custom": "cmd-value", "Authorization": "Bearer ${MY_TOKEN}"},
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(
            config,
            [cmd],
            plugin_headers={"X-Custom": "plugin-value", "Accept": "application/json"},
        )
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={})
        ctx = _mock_ctx(session)

        handler(ctx)

        headers_sent = session.request.call_args.kwargs["headers"]
        # Command header overrides plugin header
        assert headers_sent["X-Custom"] == "cmd-value"
        # Plugin header preserved
        assert headers_sent["Accept"] == "application/json"
        # Env var expanded
        assert headers_sent["Authorization"] == "Bearer abc123"


class TestHandlerQueryParams:
    """Tests for query parameter handling."""

    def test_non_url_kwargs_become_query_params(self) -> None:
        """kwargs that aren't URL params become query params."""
        cmd = _make_command(url="/users/{user_id}")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data=[])
        ctx = _mock_ctx(session)

        handler(ctx, user_id=1, page=2, limit=10)

        call_kwargs = session.request.call_args.kwargs
        assert call_kwargs["params"] == {"page": 2, "limit": 10}

    def test_none_kwargs_excluded_from_query_params(self) -> None:
        """kwargs with None values are excluded from query params."""
        cmd = _make_command(url="/items")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data=[])
        ctx = _mock_ctx(session)

        handler(ctx, page=1, filter=None)

        call_kwargs = session.request.call_args.kwargs
        assert call_kwargs["params"] == {"page": 1}


class TestHandlerResponseParsing:
    """Tests for response parsing in the handler."""

    def test_json_response_returned(self) -> None:
        """Valid JSON response is returned as parsed data."""
        cmd = _make_command(url="/data")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"items": [1, 2, 3]})
        ctx = _mock_ctx(session)

        result = handler(ctx)
        assert result == {"items": [1, 2, 3]}

    def test_non_json_response_returns_text(self) -> None:
        """When response is not JSON, returns response.text."""
        cmd = _make_command(url="/data")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(
            json_raises=True,
            text="plain text content",
            headers={"Content-Type": "text/plain"},
        )
        ctx = _mock_ctx(session)

        result = handler(ctx)
        assert result == "plain text content"

    def test_html_response_warns_user(self) -> None:
        """HTML response triggers a warning via gp_console."""
        cmd = _make_command(url="/data")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(
            json_raises=True,
            text="<!DOCTYPE html><html><body>Error</body></html>",
            headers={"Content-Type": "text/html"},
        )

        ctx = _mock_ctx(session)

        with patch("graftpunk.plugins.yaml_plugin.gp_console") as mock_console:
            result = handler(ctx)

        assert result == "<!DOCTYPE html><html><body>Error</body></html>"
        mock_console.warn.assert_called_once()
        call_arg = mock_console.warn.call_args[0][0]
        assert "Expected JSON but received HTML" in call_arg

    def test_raise_for_status_called(self) -> None:
        """When raise_for_status is True, response.raise_for_status is called."""
        cmd = _make_command(url="/data", raise_for_status=True)
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        resp = _mock_response(json_data={})
        session.request.return_value = resp
        ctx = _mock_ctx(session)

        handler(ctx)
        resp.raise_for_status.assert_called_once()

    def test_raise_for_status_not_called(self) -> None:
        """When raise_for_status is False, response.raise_for_status is NOT called."""
        cmd = _make_command(url="/data", raise_for_status=False)
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        resp = _mock_response(json_data={})
        session.request.return_value = resp
        ctx = _mock_ctx(session)

        handler(ctx)
        resp.raise_for_status.assert_not_called()


class TestHandlerJmespath:
    """Tests for jmespath filtering in the handler."""

    def test_jmespath_filter_applied(self) -> None:
        """When jmespath is set and installed, data is filtered."""
        cmd = _make_command(url="/data", jmespath="items[0]")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"items": ["first", "second"]})
        ctx = _mock_ctx(session)

        mock_jmespath = MagicMock()
        mock_jmespath.search.return_value = "first"

        with (
            patch("graftpunk.plugins.yaml_plugin.HAS_JMESPATH", True),
            patch("graftpunk.plugins.yaml_plugin._jmespath", mock_jmespath),
        ):
            result = handler(ctx)

        assert result == "first"
        mock_jmespath.search.assert_called_once_with("items[0]", {"items": ["first", "second"]})

    def test_jmespath_not_installed_warns(self) -> None:
        """When jmespath is not installed, a warning is printed and raw data returned."""
        cmd = _make_command(url="/data", jmespath="items[0]")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"items": ["first", "second"]})
        ctx = _mock_ctx(session)

        with (
            patch("graftpunk.plugins.yaml_plugin.HAS_JMESPATH", False),
            patch("graftpunk.plugins.yaml_plugin.gp_console") as mock_console,
        ):
            result = handler(ctx)

        # Returns unfiltered data
        assert result == {"items": ["first", "second"]}
        # Warning was printed
        mock_console.warn.assert_called_once()
        call_arg = mock_console.warn.call_args[0][0]
        assert "jmespath filter ignored" in call_arg

    def test_no_jmespath_filter_returns_data_as_is(self) -> None:
        """When no jmespath is specified, data is returned unfiltered."""
        cmd = _make_command(url="/data", jmespath=None)
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"items": [1, 2, 3]})
        ctx = _mock_ctx(session)

        result = handler(ctx)
        assert result == {"items": [1, 2, 3]}


class TestResourceLimitsPassthrough:
    """Tests that resource limits are passed from YAMLCommandDef to CommandSpec."""

    def test_limits_passed_to_command_spec(self) -> None:
        """Resource limit fields are forwarded to CommandSpec."""
        cmd = _make_command(
            name="limited",
            url="/api/limited",
            timeout=30.0,
            max_retries=3,
            rate_limit=1.5,
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        commands = {c.name: c for c in plugin.get_commands()}

        spec = commands["limited"]
        assert spec.timeout == 30.0
        assert spec.max_retries == 3
        assert spec.rate_limit == 1.5

    def test_default_limits_passed_to_command_spec(self) -> None:
        """Default resource limit values are forwarded to CommandSpec."""
        cmd = _make_command(name="basic", url="/api/basic")
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        commands = {c.name: c for c in plugin.get_commands()}

        spec = commands["basic"]
        assert spec.timeout is None
        assert spec.max_retries == 0
        assert spec.rate_limit is None

    def test_mixed_limits_across_commands(self) -> None:
        """Multiple commands have independent resource limits."""
        cmd_fast = _make_command(name="fast", url="/api/fast")
        cmd_slow = _make_command(
            name="slow",
            url="/api/slow",
            timeout=60.0,
            max_retries=5,
            rate_limit=0.5,
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd_fast, cmd_slow])
        commands = {c.name: c for c in plugin.get_commands()}

        assert commands["fast"].timeout is None
        assert commands["fast"].max_retries == 0
        assert commands["fast"].rate_limit is None

        assert commands["slow"].timeout == 60.0
        assert commands["slow"].max_retries == 5
        assert commands["slow"].rate_limit == 0.5


class TestConvertParamsClickKwargs:
    """Tests for _convert_params producing click_kwargs."""

    def test_string_param_converts(self) -> None:
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="query", type="str", required=True),),
        )
        params = _convert_params(cmd_def)
        assert len(params) == 1
        assert params[0].name == "query"
        assert params[0].is_option is True
        assert params[0].click_kwargs["type"] is str
        assert params[0].click_kwargs["required"] is True

    def test_int_param_converts(self) -> None:
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="limit", type="int", default=10),),
        )
        params = _convert_params(cmd_def)
        assert params[0].click_kwargs["type"] is int
        assert params[0].click_kwargs["default"] == 10

    def test_bool_param_flag_detection(self) -> None:
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="verbose", type="bool", default=False),),
        )
        params = _convert_params(cmd_def)
        assert params[0].click_kwargs.get("is_flag") is True

    def test_float_param_converts(self) -> None:
        """YAML type 'float' maps to Python float."""
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="threshold", type="float", default=0.5),),
        )
        params = _convert_params(cmd_def)
        assert params[0].click_kwargs["type"] is float
        assert params[0].click_kwargs["default"] == 0.5

    def test_help_in_click_kwargs(self) -> None:
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="q", help="Search query"),),
        )
        params = _convert_params(cmd_def)
        assert params[0].click_kwargs["help"] == "Search query"

    def test_positional_argument(self) -> None:
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test/{id}",
            params=(YAMLParamDef(name="id", is_option=False, required=True),),
        )
        params = _convert_params(cmd_def)
        assert params[0].is_option is False

    def test_help_not_in_click_kwargs_when_empty(self) -> None:
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="q"),),
        )
        params = _convert_params(cmd_def)
        assert "help" not in params[0].click_kwargs

    def test_unknown_type_defaults_to_str(self) -> None:
        """Unknown type strings in YAML params default to str and log a warning."""
        cmd_def = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            params=(YAMLParamDef(name="x", type="custom_type"),),
        )
        with patch("graftpunk.plugins.yaml_plugin.LOG") as mock_log:
            params = _convert_params(cmd_def)
        assert params[0].click_kwargs["type"] is str
        mock_log.warning.assert_called_once_with(
            "unknown_yaml_param_type",
            param="x",
            type="custom_type",
            fallback="str",
        )

    def test_yaml_help_text_in_command_spec_click_kwargs(self) -> None:
        """YAML command help_text appears in CommandSpec.click_kwargs."""
        cmd = YAMLCommandDef(name="get", help_text="Get items", method="GET", url="/items")
        config = PluginConfig(site_name="test-help", session_name="test-help", help_text="Test")
        plugin = create_yaml_site_plugin(config, [cmd])
        spec = plugin.get_commands()[0]
        assert spec.click_kwargs["help"] == "Get items"
        assert spec.help_text == "Get items"

    def test_yaml_empty_help_text_empty_click_kwargs(self) -> None:
        """YAML command with empty help_text has empty click_kwargs."""
        cmd = YAMLCommandDef(name="t", help_text="", method="GET", url="/t")
        config = PluginConfig(site_name="test-empty", session_name="test-empty", help_text="Test")
        plugin = create_yaml_site_plugin(config, [cmd])
        spec = plugin.get_commands()[0]
        assert spec.click_kwargs == {}


class TestHandlerOutputConfig:
    """Tests for output_config handling in the handler."""

    def test_output_config_returns_command_result(self) -> None:
        """When output_config is set, handler returns CommandResult."""
        from graftpunk.plugins.cli_plugin import CommandResult
        from graftpunk.plugins.output_config import ColumnFilter, OutputConfig, ViewConfig

        output_config = OutputConfig(
            views=[
                ViewConfig(
                    name="items",
                    path="results",
                    columns=ColumnFilter(mode="include", columns=["id", "name"]),
                )
            ]
        )
        cmd = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            output_config=output_config,
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(
            json_data={"results": [{"id": 1, "name": "foo"}]}
        )
        ctx = _mock_ctx(session)

        result = handler(ctx)

        assert isinstance(result, CommandResult)
        assert result.data == {"results": [{"id": 1, "name": "foo"}]}
        assert result.output_config is output_config

    def test_no_output_config_returns_raw_data(self) -> None:
        """When output_config is not set, handler returns raw data."""
        from graftpunk.plugins.cli_plugin import CommandResult

        cmd = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            output_config=None,
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"items": [1, 2, 3]})
        ctx = _mock_ctx(session)

        result = handler(ctx)

        assert not isinstance(result, CommandResult)
        assert result == {"items": [1, 2, 3]}

    def test_output_config_with_jmespath(self) -> None:
        """When both jmespath and output_config are set, jmespath is applied first."""
        from graftpunk.plugins.cli_plugin import CommandResult
        from graftpunk.plugins.output_config import ColumnFilter, OutputConfig, ViewConfig

        output_config = OutputConfig(
            views=[ViewConfig(name="items", columns=ColumnFilter(mode="include", columns=["id"]))]
        )
        cmd = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            jmespath="data.items",
            output_config=output_config,
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(
            json_data={"data": {"items": [{"id": 1}, {"id": 2}]}}
        )
        ctx = _mock_ctx(session)

        mock_jmespath = MagicMock()
        mock_jmespath.search.return_value = [{"id": 1}, {"id": 2}]

        with (
            patch("graftpunk.plugins.yaml_plugin.HAS_JMESPATH", True),
            patch("graftpunk.plugins.yaml_plugin._jmespath", mock_jmespath),
        ):
            result = handler(ctx)

        assert isinstance(result, CommandResult)
        assert result.data == [{"id": 1}, {"id": 2}]
        assert result.output_config is output_config

    def test_output_config_with_jmespath_not_installed(self) -> None:
        """When jmespath not installed, returns CommandResult with raw data."""
        from graftpunk.plugins.cli_plugin import CommandResult
        from graftpunk.plugins.output_config import ColumnFilter, OutputConfig, ViewConfig

        output_config = OutputConfig(
            views=[ViewConfig(name="items", columns=ColumnFilter(mode="include", columns=["id"]))]
        )
        cmd = YAMLCommandDef(
            name="test",
            help_text="",
            method="GET",
            url="/test",
            jmespath="data.items",
            output_config=output_config,
        )
        config = _make_config()
        plugin = create_yaml_site_plugin(config, [cmd])
        handler = {c.name: c for c in plugin.get_commands()}["test"].handler

        session = MagicMock(spec=requests.Session)
        session.request.return_value = _mock_response(json_data={"data": {"items": [{"id": 1}]}})
        ctx = _mock_ctx(session)

        with (
            patch("graftpunk.plugins.yaml_plugin.HAS_JMESPATH", False),
            patch("graftpunk.plugins.yaml_plugin.gp_console"),
        ):
            result = handler(ctx)

        # Even though jmespath couldn't be applied, output_config still wraps result
        assert isinstance(result, CommandResult)
        assert result.data == {"data": {"items": [{"id": 1}]}}
        assert result.output_config is output_config
